import boto3
from cloudlift.exceptions import UnrecoverableException

from cloudlift.config import EnvironmentConfiguration
from cloudlift.config.logging import log_err

def get_region_for_environment(environment):
    if environment:
        return EnvironmentConfiguration(environment).get_config()[environment]['region']
    else:
        # Get the region from the AWS credentials used to execute cloudlift
        aws_session = boto3.session.Session()
        return aws_session.region_name


def get_client_for(resource, environment):
    return boto3.session.Session(
        region_name=get_region_for_environment(environment)
    ).client(resource)


def get_resource_for(resource, environment):
    return boto3.session.Session(
        region_name=get_region_for_environment(environment)
    ).resource(resource)


def get_notifications_arn_for_environment(environment):
    try:
        return EnvironmentConfiguration(
            environment
        ).get_config()[environment]['environment']["notifications_arn"]
    except KeyError:
        raise UnrecoverableException("Unable to find notifications arn for {environment}".format(**locals()))


def get_ssl_certification_for_environment(environment):
    try:
        return EnvironmentConfiguration(
            environment
        ).get_config()[environment]['environment']["ssl_certificate_arn"]
    except KeyError:
        raise UnrecoverableException("Unable to find ssl certificate for {environment}".format(**locals()))



from cloudlift.config.logging import log_intent, log_intent_err


def get_stack_events(client, stack_name):
    try:
        return sorted(
            client.describe_stack_events(
                StackName=stack_name
            )['StackEvents'], key=lambda k: k['Timestamp'])
    except Exception:
        return []


def print_new_events(all_events, existing_events):
    new_events = [evnt for evnt in all_events if evnt not in existing_events]
    for event in new_events:
        update = "%s: Resource: %s\t\tStatus: %s" % (
            event['Timestamp'],
            event['LogicalResourceId'],
            event['ResourceStatus']
        )
        if 'ResourceStatusReason' in event:
            update += "\t\tReason: %s" % event['ResourceStatusReason']
        if "ERROR" in update or "FAIL" in update:
            log_intent_err(update)
        else:
            log_intent(update)



import click


def log_err(text):
    click.secho(text, fg='red', bold=True)

def log(text):
    click.secho(text, fg='green')

def log_warning(text):
    click.secho(text, fg='yellow')

def log_bold(text):
    click.secho(text, fg='green',  bold=True)

def log_intent(text, level=1):
    click.secho(''.join(['  '] * level + [text]), fg='green')

def log_intent_err(text, level=1):
    click.secho(''.join(['  '] * level + [text]), fg='red')

def log_with_color(text, color, level=2):
    click.secho(''.join(['  '] * level + [text]), fg=color)



import re

from cloudlift.exceptions import UnrecoverableException

from cloudlift.config import get_client_for
from cloudlift.config.logging import log_err


class ParameterStore(object):
    def __init__(self, service_name, environment):
        self.service_name = service_name
        self.environment = environment
        self.path_prefix = "/%s/%s/" % (self.environment, self.service_name)
        # TODO: Use the below two lines when all parameter store actions
        # require MFA
        #
        # mfa_region = get_region_for_environment(environment)
        # mfa_session = mfa.get_mfa_session(mfa_region)
        # ssm_client = mfa_session.client('ssm')
        self.client = get_client_for('ssm', environment)

    def get_existing_config_as_string(self):
        environment_configs = self.get_existing_config()
        return '\n'.join('{}={}'.format(key, val) for key, val in sorted(
            environment_configs.items()
        ))

    def get_existing_config(self):
        environment_configs = {}
        next_token = None
        while True:
            if next_token:
                response = self.client.get_parameters_by_path(
                    Path=self.path_prefix,
                    Recursive=False,
                    WithDecryption=True,
                    MaxResults=10,
                    NextToken=next_token
                )
            else:
                response = self.client.get_parameters_by_path(
                    Path=self.path_prefix,
                    Recursive=False,
                    WithDecryption=True,
                    MaxResults=10
                )
            for parameter in response['Parameters']:
                parameter_name = parameter['Name'].split(self.path_prefix)[1]
                environment_configs[parameter_name] = parameter['Value']

            try:
                next_token = response['NextToken']
            except KeyError:
                break
        return environment_configs

    def set_config(self, differences):
        self._validate_changes(differences)
        for parameter_change in differences:
            if parameter_change[0] == 'change':
                response = self.client.put_parameter(
                    Name='%s%s' % (self.path_prefix, parameter_change[1]),
                    Value=parameter_change[2][1],
                    Type='SecureString',
                    KeyId='alias/aws/ssm',
                    Overwrite=True
                )
            elif parameter_change[0] == 'add':
                for added_parameter in parameter_change[2]:
                    response = self.client.put_parameter(
                        Name='%s%s' % (self.path_prefix, added_parameter[0]),
                        Value=added_parameter[1],
                        Type='SecureString',
                        KeyId='alias/aws/ssm',
                        Overwrite=False
                    )
            elif parameter_change[0] == 'remove':
                deleted_parameters = ["%s%s" % (self.path_prefix, item[0]) for item in parameter_change[2]]
                response = self.client.delete_parameters(
                    Names=deleted_parameters
                )

    def _validate_changes(self, differences):
        errors = []
        for parameter_change in differences:
            if parameter_change[0] == 'change':
                if not self._is_a_valid_parameter_key(parameter_change[1]):
                    errors.append("'%s' is not a valid key." % parameter_change[1])
            elif parameter_change[0] == 'add':
                for added_parameter in parameter_change[2]:
                    if not self._is_a_valid_parameter_key(added_parameter[0]):
                        errors.append("'%s' is not a valid key." % added_parameter[0])
            elif parameter_change[0] == 'remove':
                # No validation required
                pass
        if errors:
            for error in errors:
                log_err(error)
            raise UnrecoverableException("Environment variables validation failed with above errors.")
        return True

    def _is_a_valid_parameter_key(self, key):
        return bool(re.match(r"^[\w|\.|\-|\/]+$", key))




import click
import dictdiffer

from cloudlift.config import print_parameter_changes
from cloudlift.config import ParameterStore
from cloudlift.deployment.deployer import read_config
from cloudlift.config.logging import log_warning


def edit_config(name, environment):
    parameter_store = ParameterStore(name, environment)
    env_config_strings = parameter_store.get_existing_config_as_string()
    edited_config_content = click.edit(str(env_config_strings))

    if edited_config_content is None:
        log_warning("No changes made, exiting.")
        return

    differences = list(dictdiffer.diff(
        read_config(env_config_strings),
        read_config(edited_config_content)
    ))
    if not differences:
        log_warning("No changes made, exiting.")
    else:
        print_parameter_changes(differences)
        if click.confirm('Do you want update the config?'):
            parameter_store.set_config(differences)
        else:
            log_warning("Changes aborted.")



import sys
import uuid
from time import sleep

import click

from cloudlift.config.logging import log, log_bold, log_err

def create_change_set(client, service_template_body, template_source, stack_name,
                      key_name, environment):
    change_set_parameters = [
        {'ParameterKey': 'Environment', 'ParameterValue': environment}
    ]
    if key_name:
        change_set_parameters.append({
            'ParameterKey': 'KeyPair',
            'ParameterValue': key_name
        })
    if template_source == 'TemplateBody':
        create_change_set_res = client.create_change_set(
            StackName=stack_name,
            ChangeSetName="cg"+uuid.uuid4().hex,
            TemplateBody=service_template_body,
            Parameters=change_set_parameters,
            Capabilities=['CAPABILITY_NAMED_IAM'],
            ChangeSetType='UPDATE'
        )
    elif template_source == 'TemplateURL':
        create_change_set_res = client.create_change_set(
            StackName=stack_name,
            ChangeSetName="cg"+uuid.uuid4().hex,
            TemplateURL=service_template_body,
            Parameters=change_set_parameters,
            Capabilities=['CAPABILITY_NAMED_IAM'],
            ChangeSetType='UPDATE'
        )
    log("Changeset creation initiated. Checking the progress...")
    change_set = client.describe_change_set(
        ChangeSetName=create_change_set_res['Id']
    )
    while change_set['Status'] in ['CREATE_PENDING', 'CREATE_IN_PROGRESS']:
        sleep(1)
        status_string = '\x1b[2K\rChecking changeset status.  Status: ' + \
                        change_set['Status']
        sys.stdout.write(status_string)
        sys.stdout.flush()
        change_set = client.describe_change_set(
            ChangeSetName=create_change_set_res['Id']
        )
    status_string = '\x1b[2K\rChecking changeset status..  Status: ' + \
                    change_set['Status']+'\n'
    sys.stdout.write(status_string)
    if change_set['Status'] == 'FAILED':
        log_err("Changeset creation failed!")
        log_bold(change_set.get(
            'StatusReason',
            "Check AWS console for reason."
        ))
        client.delete_change_set(ChangeSetName=create_change_set_res['Id'])
    else:
        log_bold("Changeset created.. Following are the changes")
        _print_changes(change_set)
        if click.confirm('Do you want to execute the changeset?'):
            return change_set
        log_bold("Deleting changeset...")
        client.delete_change_set(
            ChangeSetName=create_change_set_res['Id']
        )
        log_bold("Done. Bye!")


def _print_changes(change_set):
    for change in change_set['Changes']:
        resource_change = change['ResourceChange']
        change_line = click.style(
            resource_change['Action'] + ": " +
            resource_change['LogicalResourceId'] +
            " (" + resource_change['ResourceType'] + "/" +
            resource_change.get('PhysicalResourceId', '--') + ")\n",
            fg='green', bold=True) + \
            click.style("  "+str(resource_change['Details']), fg='green')
        click.echo(change_line)



import sys
from time import sleep

from botocore.exceptions import ClientError
from cloudlift.exceptions import UnrecoverableException

from cloudlift.config import EnvironmentConfiguration
from cloudlift.config import get_client_for
from cloudlift.config import get_cluster_name
from cloudlift.deployment.changesets import create_change_set
from cloudlift.deployment.cluster_template_generator import ClusterTemplateGenerator
from cloudlift.config.logging import log, log_bold, log_err
from cloudlift.deployment.progress import get_stack_events, print_new_events


class EnvironmentCreator(object):

    def __init__(self, environment):
        self.environment = environment
        environment_configuration = EnvironmentConfiguration(
            self.environment
        )
        environment_configuration.update_config()
        self.configuration = environment_configuration.get_config()[self.environment]
        self.cluster_name = get_cluster_name(environment)
        self.client = get_client_for('cloudformation', self.environment)
        self.key_name = self.configuration['cluster']['key_name']

    def run(self):
        try:
            log("Check if stack already exists for " + self.cluster_name)
            environment_stack = self.client.describe_stacks(
                StackName=self.cluster_name
            )['Stacks'][0]
            log(self.cluster_name + " stack exists. ID: " +
                environment_stack['StackId'])
            log_err("Cannot create environment with duplicate name: " +
                    self.cluster_name)
        except Exception:
            log(self.cluster_name +
                " stack does not exist. Creating new stack.")
            # When creating a cluster, desired_instance count is same
            # as min_instance count
            environment_stack_template_body = ClusterTemplateGenerator(
                self.environment,
                self.configuration
            ).generate_cluster()
            self.existing_events = get_stack_events(
                self.client,
                self.cluster_name
            )
            environment_stack = self.client.create_stack(
                StackName=self.cluster_name,
                TemplateBody=environment_stack_template_body,
                Parameters=[
                    {
                        'ParameterKey': 'KeyPair',
                        'ParameterValue': self.key_name,
                    },
                    {
                        'ParameterKey': 'Environment',
                        'ParameterValue': self.environment,
                    }
                ],
                OnFailure='DO_NOTHING',
                Capabilities=['CAPABILITY_NAMED_IAM'],
            )
            log_bold("Submitted to cloudformation. Checking progress...")
            self.__print_progress()
            log_bold(self.cluster_name+" stack created. ID: " +
                     environment_stack['StackId'])

    def run_update(self, update_ecs_agents):
        if update_ecs_agents:
            self.__run_ecs_container_agent_udpate()
        try:
            log("Initiating environment stack update.")
            environment_stack_template_body = ClusterTemplateGenerator(
                self.environment,
                self.configuration,
                self.__get_desired_count()
            ).generate_cluster()
            log("Template generation complete.")
            change_set = create_change_set(
                self.client,
                environment_stack_template_body,
                self.cluster_name,
                self.key_name,
                self.environment
            )
            self.existing_events = get_stack_events(
                self.client,
                self.cluster_name
            )
            log_bold("Executing changeset. Checking progress...")
            self.client.execute_change_set(
                ChangeSetName=change_set['ChangeSetId']
            )
            self.__print_progress()
        except ClientError as e:
            log_err("No updates are to be performed")
        except Exception as e:
            raise e
            # log_err("Something went wrong")
            # TODO: Describe error here.

    def __get_desired_count(self):
        try:
            auto_scaling_client = get_client_for(
                'autoscaling',
                self.environment
            )
            cloudformation_client = get_client_for(
                'cloudformation',
                self.environment
            )
            cfn_resources = cloudformation_client.list_stack_resources(
                StackName=self.cluster_name
            )
            auto_scaling_group_name = list(
                filter(
                    lambda x: x['ResourceType'] == "AWS::AutoScaling::AutoScalingGroup",
                    cfn_resources['StackResourceSummaries']
                )
            )[0]['PhysicalResourceId']
            response = auto_scaling_client.describe_auto_scaling_groups(
                AutoScalingGroupNames=[auto_scaling_group_name]
            )
            return response['AutoScalingGroups'][0]['DesiredCapacity']
        except Exception:
            raise UnrecoverableException("Unable to fetch desired instance count.")


    def __print_progress(self):
        while True:
            response = self.client.describe_stacks(StackName=self.cluster_name)
            if "IN_PROGRESS" not in response['Stacks'][0]['StackStatus']:
                break
            all_events = get_stack_events(self.client, self.cluster_name)
            print_new_events(all_events, self.existing_events)
            self.existing_events = all_events
            sleep(5)
        log_bold("Finished and Status: %s" % (response['Stacks'][0]['StackStatus']))

    def __run_ecs_container_agent_udpate(self):
        log("Initiating agent update")
        ecs_client = get_client_for('ecs', self.environment)
        response = ecs_client.list_container_instances(
            cluster=self.cluster_name
        )
        container_instance_arns = response['containerInstanceArns']
        for container_instance_arn in response['containerInstanceArns']:
            try:
                response = ecs_client.update_container_agent(
                    cluster=self.cluster_name,
                    containerInstance=container_instance_arn
                )
            except ClientError as exception:
                if "There is no update available for your container agent." in str(exception):
                    log("There is no update available for your container agent " + container_instance_arn)
                elif "Agent update is already in progress." in str(exception):
                    log("Agent update is already in progress " + container_instance_arn)
                else:
                    raise exception

        while True:
            sleep(1)
            response = ecs_client.describe_container_instances(
                cluster=self.cluster_name,
                containerInstances=container_instance_arns
            )
            update_statuses = map(
                lambda x: {
                    "arn": x['containerInstanceArn'],
                    "status": x.get('agentUpdateStatus', 'UPDATED')
                },
                response['containerInstances']
            )
            finished = True
            status_string = '\r'
            for status in update_statuses:
                status_string += status['arn'] + ":\033[92m" + \
                    status['status'] + " \033[0m"
                if status['status'] != 'UPDATED':
                    finished = False
            sys.stdout.write(status_string)
            sys.stdout.flush()

            if finished:
                print("")
                break



'''
This module abstracts implementation of storing, editing and
retrieving service configuration.
'''

import json

import dictdiffer
from botocore.exceptions import ClientError
from click import confirm, edit
from cloudlift.exceptions import UnrecoverableException
from jsonschema import validate
from jsonschema.exceptions import ValidationError
from stringcase import pascalcase

from cloudlift.config import DecimalEncoder
from cloudlift.config import print_json_changes
# import config.mfa as mfa
from cloudlift.config import get_resource_for
from cloudlift.config.logging import log_bold, log_err, log_warning
from cloudlift.version import VERSION

SERVICE_CONFIGURATION_TABLE = 'service_configurations'

class ServiceConfiguration(object):
    '''
        Handles configuration in DynamoDB for services
    '''

    def __init__(self, service_name, environment):
        self.service_name = service_name
        self.environment = environment
        self.new_service = False
        # TODO: Use the below two lines when all parameter store actions
        # require MFA
        #
        # mfa_region = get_region_for_environment(environment)
        # mfa_session = mfa.get_mfa_session(mfa_region)
        # ssm_client = mfa_session.client('ssm')
        self.table = get_resource_for(
            'dynamodb',
            environment
        ).Table(SERVICE_CONFIGURATION_TABLE)

    def edit_config(self):
        '''
            Open editor to update configuration
        '''

        try:
            current_configuration = self.get_config()

            updated_configuration = edit(
                json.dumps(
                    current_configuration,
                    indent=4,
                    sort_keys=True,
                    cls=DecimalEncoder
                )
            )

            if updated_configuration is None:
                if self.new_service:
                    self.set_config(current_configuration)
                    log_warning("Using default configuration.")
                else:
                    log_warning("No changes made.")
            else:
                updated_configuration = json.loads(updated_configuration)
                differences = list(dictdiffer.diff(
                    current_configuration,
                    updated_configuration
                ))
                if not differences:
                    log_warning("No changes made.")
                else:
                    print_json_changes(differences)
                    if confirm('Do you want update the config?'):
                        self.set_config(updated_configuration)
                    else:
                        log_warning("Changes aborted.")
        except ClientError:
            raise UnrecoverableException("Unable to fetch service configuration from DynamoDB.")

    def get_config(self):
        '''
            Get configuration from DynamoDB
        '''

        try:
            configuration_response = self.table.get_item(
                Key={
                    'service_name': self.service_name,
                    'environment': self.environment
                },
                ConsistentRead=True,
                AttributesToGet=[
                    'configuration'
                ]
            )
            if 'Item' in configuration_response:
                existing_configuration = configuration_response['Item']['configuration']
            else:
                existing_configuration = self._default_service_configuration()
                self.new_service = True

            existing_configuration.pop("cloudlift_version", None)
            return existing_configuration
        except ClientError:
            raise UnrecoverableException("Unable to fetch service configuration from DynamoDB.")

    def set_config(self, config):
        '''
            Set configuration in DynamoDB
        '''
        config['cloudlift_version'] = VERSION
        self._validate_changes(config)
        try:
            configuration_response = self.table.update_item(
                TableName=SERVICE_CONFIGURATION_TABLE,
                Key={
                    'service_name': self.service_name,
                    'environment': self.environment
                },
                UpdateExpression='SET configuration = :configuration',
                ExpressionAttributeValues={
                    ':configuration': config
                },
                ReturnValues="UPDATED_NEW"
            )
            return configuration_response
        except ClientError:
            raise UnrecoverableException("Unable to store service configuration in DynamoDB.")

    def update_cloudlift_version(self):
        '''
            Updates cloudlift version in service configuration
        '''
        config = self.get_config()
        self.set_config(config)

    def _validate_changes(self, configuration):
        service_schema = {
            "title": "service",
            "type": "object",
            "properties": {
                "http_interface": {
                    "type": "object",
                    "properties": {
                        "internal": {
                            "type": "boolean"
                        },
                        "restrict_access_to": {
                            "type": "array",
                            "items": {
                                "type": "string"
                            }
                        },
                        "container_port": {
                            "type": "number"
                        },
                        "health_check_path": {
                            "type": "string",
                            "pattern": "^\/.*$"
                        }
                    },
                    "required": [
                        "internal",
                        "restrict_access_to",
                        "container_port"
                    ]
                },
                "memory_reservation": {
                    "type": "number",
                    "minimum": 10,
                    "maximum": 30000
                },
                "fargate": {
                    "type": "object",
                    "properties": {
                        "cpu": {
                            "type": "number",
                            "minimum": 256,
                            "maximum": 4096
                        },
                        "memory": {
                            "type": "number",
                            "minimum": 512,
                            "maximum": 30720
                        }
                    }
                },
                "command": {
                    "oneOf": [
                        {"type": "string"},
                        {"type": "null"}
                    ]
                }
            },
            "required": ["memory_reservation", "command"]
        }
        schema = {
            # "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "configuration",
            "type": "object",
            "properties": {
                "notifications_arn": {
                    "type": "string"
                },
                "services": {
                    "type": "object",
                    "patternProperties": {
                        "^[a-zA-Z]+$": service_schema
                    }
                },
                "cloudlift_version": {
                    "type": "string"
                }
            },
            "required": ["cloudlift_version", "services"]
        }
        try:
            validate(configuration, schema)
        except ValidationError as validation_error:
            raise UnrecoverableException(validation_error.message + " in " +
                    str(".".join(list(validation_error.relative_path))))
        log_bold("Schema valid!")

    def _default_service_configuration(self):
        return {
            u'services': {
                pascalcase(self.service_name): {
                    u'http_interface': {
                        u'internal': False,
                        u'restrict_access_to': [u'0.0.0.0/0'],
                        u'container_port': 80,
                        u'health_check_path': u'/elb-check'
                    },
                    u'memory_reservation': 1000,
                    u'command': None
                }
            }
        }



from time import sleep

from colorclass import Color
from terminaltables import SingleTable

from cloudlift.config import ParameterStore
from cloudlift.config.logging import log_bold, log_err, log_intent, log_with_color
from cloudlift.deployment.ecs import DeployAction, EcsClient
from cloudlift.exceptions import UnrecoverableException


def deploy_new_version(region, cluster_name, ecs_service_name,
                       deploy_version_tag, service_name, sample_env_file_path,
                       env_name, color='white', complete_image_uri=None):
    env_config = build_config(env_name, service_name, sample_env_file_path)
    client = EcsClient(None, None, region)
    deployment = DeployAction(client, cluster_name, ecs_service_name)
    if deployment.service.desired_count == 0:
        desired_count = 1
    else:
        desired_count = deployment.service.desired_count
    deployment.service.set_desired_count(desired_count)
    task_definition = deployment.get_current_task_definition(
        deployment.service
    )
    if complete_image_uri is not None:
        container_name = task_definition['containerDefinitions'][0]['name']
        task_definition.set_images(
            deploy_version_tag,
            **{container_name: complete_image_uri}
        )
    else:
        task_definition.set_images(deploy_version_tag)
    for container in task_definition.containers:
        task_definition.apply_container_environment(container, env_config)
    print_task_diff(ecs_service_name, task_definition.diff, color)
    new_task_definition = deployment.update_task_definition(task_definition)
    response = deploy_and_wait(deployment, new_task_definition, color)
    if response:
        log_bold(ecs_service_name + " Deployed successfully.")
    else:
        log_err(ecs_service_name + " Deployment failed.")
    return response


def deploy_and_wait(deployment, new_task_definition, color):
    existing_events = fetch_events(deployment.get_service())
    deployment.deploy(new_task_definition)
    return wait_for_finish(deployment, existing_events, color)


def build_config(env_name, service_name, sample_env_file_path):
    service_config = read_config(open(sample_env_file_path).read())
    try:
        environment_config = ParameterStore(
            service_name,
            env_name).get_existing_config()
    except Exception as err:
        log_intent(str(err))
        raise UnrecoverableException("Cannot find the configuration in parameter store \
[env: %s | service: %s]." % (env_name, service_name))
    missing_env_config = set(service_config) - set(environment_config)
    if missing_env_config:
        raise UnrecoverableException('There is no config value for the keys ' +
                str(missing_env_config))
    missing_env_sample_config = set(environment_config) - set(service_config)
    if missing_env_sample_config:
        raise UnrecoverableException('There is no config value for the keys in env.sample file ' +
                str(missing_env_sample_config))

    return make_container_defn_env_conf(service_config, environment_config)


def read_config(file_content):
    config = {}
    for line in file_content.splitlines():
        line = line.strip()
        if not line:
            continue
        key, value = line.split('=', 1)
        config[key] = value
    return config


def make_container_defn_env_conf(service_config, environment_config):
    container_defn_env_config = []
    for env_var_name in service_config:
        container_defn_env_config.append(
            (env_var_name, environment_config[env_var_name])
        )
    return container_defn_env_config


def wait_for_finish(action, existing_events, color):
    waiting = True
    while waiting:
        sleep(1)
        service = action.get_service()
        existing_events = fetch_and_print_new_events(
            service,
            existing_events,
            color
        )
        waiting = not action.is_deployed(service) and not service.errors
    if service.errors:
        log_err(str(service.errors))
        return False
    return True


def fetch_events(service):
    all_events = sorted(service.get(u'events'), key=lambda k: k['createdAt'])
    return all_events


def fetch_and_print_new_events(service, existing_events, color):
    all_events = fetch_events(service)
    new_events = [evnt for evnt in all_events if evnt not in existing_events]
    for event in new_events:
        log_with_color(
            event['message'].replace("(", "").replace(")", "")[8:],
            color
        )
    return all_events


def print_task_diff(ecs_service_name, diffs, color):
    image_diff = next(x for x in diffs if x.field == 'image')
    if image_diff.old_value != image_diff.value:
        log_with_color(ecs_service_name + " New image getting deployed", color)
        log_with_color(ecs_service_name + " " + str(image_diff), color)
    else:
        log_with_color(ecs_service_name + " No change in image version", color)
    env_diff = next(x for x in diffs if x.field == 'environment')
    old_env, current_env = env_diff.old_value, env_diff.value
    env_vars = sorted(
        set(env_diff.old_value.keys()).union(env_diff.value.keys())
    )
    table_data = []
    table_data.append(
        [
            Color('{autoyellow}Env. var.{/autoyellow}'),
            Color('{autoyellow}Old value{/autoyellow}'),
            Color('{autoyellow}Current value{/autoyellow}')
        ]
    )
    for env_var in env_vars:
        old_val = old_env.get(env_var, '-')
        current_val = current_env.get(env_var, '-')
        if old_val != current_val:
            env_var_diff_color = 'autored'
            table_data.append(
                [
                    Color(
                        '{' + env_var_diff_color + '}' +
                        env_var +
                        '{/'+env_var_diff_color+'}'
                    ),
                    old_val,
                    current_val
                ]
            )
    if len(table_data) > 1:
        log_with_color(ecs_service_name + " Environment changes", color)
        print(SingleTable(table_data).table)
    else:
        log_with_color(
            ecs_service_name + " No change in environment variables",
            color
        )



from troposphere import Output, Ref, Template

from cloudlift.config import region as region_service
from cloudlift.config import get_cluster_name


class TemplateGenerator(object):
    """This is the base class for all templates"""

    def __init__(self, env):
        self.template = Template()
        self.env = env
        self.cluster_name = get_cluster_name(env)

    def _add_stack_outputs(self):
        self.template.add_output(
            Output(
                "StackId",
                Description="The unique ID of the stack. To be supplied to \
circle CI environment variables to validate during deployment.",
                Value=Ref("AWS::StackId")
            )
        )
        self.template.add_output(
            Output(
                "StackName",
                Description="The name of the stack",
                Value=Ref('AWS::StackName')
            )
        )

    @property
    def region(self):
        return region_service.get_region_for_environment(self.env)

    @property
    def notifications_arn(self):
        return region_service.get_notifications_arn_for_environment(self.env)

    @property
    def ssl_certificate_arn(self):
        return region_service.get_ssl_certification_for_environment(self.env)



import functools
import operator

from cloudlift.exceptions import UnrecoverableException

from cloudlift.config import get_client_for, get_region_for_environment
from cloudlift.config import mfa
from cloudlift.config.logging import log, log_bold, log_err
from cloudlift.deployment.service_information_fetcher import ServiceInformationFetcher
from awscli.clidriver import create_clidriver

class SessionCreator(object):
  def __init__(self, name, environment):
    self.name = name
    self.environment = environment
    self.sts_client = get_client_for("sts", self.environment)

  def start_session(self, mfa_code):
    mfa.do_mfa_login(mfa_code, get_region_for_environment(self.environment))
    target_instance = self._get_target_instance()
    self._initiate_session(target_instance)

  def _get_target_instance(self):
    service_instance_ids = ServiceInformationFetcher(self.name, self.environment).get_instance_ids()
    if not service_instance_ids:
      raise UnrecoverableException("Couldn't find instances. Exiting.")
    instance_ids = list(set(functools.reduce(operator.add,service_instance_ids.values())))
    log("Found " + str(len(instance_ids)) + " instances to start session")
    return instance_ids[0]

  def _initiate_session(self, target_instance):
    log_bold("Starting session in " + target_instance)
    try:
      driver = create_clidriver()
      driver.main(["ssm", "start-session", "--target", target_instance])
    except:
      raise UnrecoverableException("Failed to start session")



from .changesets import *
from .cluster_template_generator import *
from .configs import *
from .deployer import *
from .ecs import *
from .environment_creator import *
from .progress import *
from .service_creator import *
from .service_information_fetcher import *
from .service_template_generator import *
from .service_updater import *
from .template_generator import *


import base64
import subprocess

import boto3
from cloudlift.exceptions import UnrecoverableException
from stringcase import spinalcase


from cloudlift.config import get_account_id
from cloudlift.config.logging import log_intent, log_warning, log_bold, log_err


class EcrClient:
    def __init__(self, name, region, build_args=None, working_dir='.'):
        self.name = name
        self.build_args = build_args
        self.working_dir = working_dir
        self.region = region
        self.ecr_client = boto3.session.Session(region_name=self.region).client('ecr')

    def build_and_upload_image(self):
        self._ensure_repository()
        self._ensure_image_in_ecr()

    def upload_image(self, version, additional_tags):
        image_name = spinalcase(self.name) + ':' + version
        ecr_image_name = self.ecr_image_uri + ':' + version
        self._ensure_repository()
        self._push_image(image_name, ecr_image_name)

        for new_tag in additional_tags:
            self._add_image_tag(version, new_tag)

    def _ensure_repository(self):
        try:
            self.ecr_client.create_repository(
                repositoryName=self.repo_name,
                imageScanningConfiguration={
                    'scanOnPush': True
                },
            )
            log_intent('Repo created with name: ' + self.repo_name)
        except Exception as ex:
            if type(ex).__name__ == 'RepositoryAlreadyExistsException':
                log_intent('Repo exists with name: ' + self.repo_name)
            else:
                raise ex

    def _ensure_image_in_ecr(self):
        if self.version == 'dirty':
            image = None
        else:
            image = self._find_image_in_ecr(self.version)
        if image:
            log_intent("Image found in ECR")
        else:
            log_bold(f"Image not found in ECR. Building image {self.name} {self.version}")
            image_name = spinalcase(self.name) + ':' + self.version
            ecr_name = self.ecr_image_uri + ':' + self.version
            self._build_image(image_name)
            self._push_image(image_name, ecr_name)
            image = self._find_image_in_ecr(self.version)

        try:
            image_manifest = image['imageManifest']
            self.ecr_client.put_image(
                repositoryName=self.repo_name,
                imageTag=self.version,
                imageManifest=image_manifest
            )
        except Exception:
            pass

    def set_version(self, version):
        if version:
            try:
                commit_sha = self._find_commit_sha(version)
            except:
                commit_sha = version
            log_intent("Using commit hash " + commit_sha + " to find image")
            image = self._find_image_in_ecr(commit_sha)
            if not image:
                log_warning("Please build, tag and upload the image for the \
        commit " + commit_sha)
                raise UnrecoverableException("Image for given version could not be found.")
            self.version = version
        else:
            dirty = subprocess.check_output(
                ["git", "status", "--short"]
            ).decode("utf-8")
            if dirty:
                self.version = 'dirty'
                log_intent("Version parameter was not provided. Determined \
        version to be " + self.version + " based on current status")
            else:
                self.version = self._find_commit_sha()
                log_intent("Version parameter was not provided. Determined \
        version to be " + self.version + " based on current status")

    def _build_image(self, image_name):
        log_bold("Building docker image " + image_name)
        command = self._build_command(image_name)
        subprocess.check_call(command, shell=True)
        log_bold("Built " + image_name)

    def _build_command(self, image_name):
        if self.build_args is None:
            return f'docker build -t {image_name} {self.working_dir}'
        else:
            build_args_command_fragment = []
            for k, v in self.build_args.items():
                build_args_command_fragment.append(" --build-arg " + "=".join((k, v)))
            return f'docker build -t {image_name}{"".join(build_args_command_fragment)} {self.working_dir}'

    def _login_to_ecr(self):
        log_intent("Attempting login...")
        auth_token_res = self.ecr_client.get_authorization_token()
        user, auth_token = base64.b64decode(
            auth_token_res['authorizationData'][0]['authorizationToken']
        ).decode("utf-8").split(':')
        ecr_url = auth_token_res['authorizationData'][0]['proxyEndpoint']
        subprocess.check_call(["docker", "login", "-u", user,
                               "-p", auth_token, ecr_url])
        log_intent('Docker login to ECR succeeded.')

    def _find_commit_sha(self, version=None):
        log_intent("Finding commit SHA")
        try:
            version_to_find = version or "HEAD"
            commit_sha = subprocess.check_output(
                ["git", "rev-list", "-n", "1", version_to_find]
            ).strip().decode("utf-8")
            log_intent("Found commit SHA " + commit_sha)
            return commit_sha
        except:
            raise UnrecoverableException("Commit SHA not found. Given version is not a git tag, \
branch or commit SHA")

    def _push_image(self, local_name, ecr_name):
        try:
            subprocess.check_call(["docker", "tag", local_name, ecr_name])
        except:
            raise UnrecoverableException("Local image was not found.")
        self._login_to_ecr()
        subprocess.check_call(["docker", "push", ecr_name])
        subprocess.check_call(["docker", "rmi", ecr_name])
        log_intent('Pushed the image (' + local_name + ') to ECR sucessfully.')

    def _add_image_tag(self, existing_tag, new_tag):
        try:
            image_manifest = self.ecr_client.batch_get_image(
                repositoryName=self.repo_name,
                imageIds=[
                    {'imageTag': existing_tag}
                ])['images'][0]['imageManifest']
            self.ecr_client.put_image(
                repositoryName=self.repo_name,
                imageTag=new_tag,
                imageManifest=image_manifest
            )
        except:
            log_err("Unable to add additional tag " + str(new_tag))

    def _find_image_in_ecr(self, tag):
        try:
            return self.ecr_client.batch_get_image(
                repositoryName=self.repo_name,
                imageIds=[{'imageTag': tag}]
            )['images'][0]
        except:
            return None

    @property
    def repo_name(self):
        return self.name + '-repo'

    @property
    def ecr_image_uri(self):
        return str(self.account_id) + ".dkr.ecr." + self.region + \
               ".amazonaws.com/" + self.repo_name

    @property
    def account_id(self):
        return get_account_id()



import json
import re

from cfn_flip import to_yaml
from stringcase import camelcase, pascalcase
from troposphere import (Base64, FindInMap, Output, Parameter, Ref, Sub,
                         cloudformation)
from troposphere.autoscaling import (AutoScalingGroup, LaunchConfiguration,
                                     ScalingPolicy)
from troposphere.cloudwatch import Alarm, MetricDimension
from troposphere.ec2 import (VPC, InternetGateway, NatGateway, Route,
                             RouteTable, SecurityGroup, Subnet,
                             SubnetRouteTableAssociation, VPCGatewayAttachment)
from troposphere.ecs import Cluster
from troposphere.elasticache import SubnetGroup as ElastiCacheSubnetGroup
from troposphere.iam import InstanceProfile, Role
from troposphere.logs import LogGroup
from troposphere.policies import (AutoScalingRollingUpdate, CreationPolicy,
                                  ResourceSignal)
from troposphere.rds import DBSubnetGroup

from cloudlift.config import DecimalEncoder
from cloudlift.config import get_client_for, get_region_for_environment
from cloudlift.deployment.template_generator import TemplateGenerator
from cloudlift.version import VERSION


class ClusterTemplateGenerator(TemplateGenerator):
    """
        This class generates CloudFormation template for a environment cluster
    """

    def __init__(self, environment, environment_configuration, desired_instances=None):
        super(ClusterTemplateGenerator, self).__init__(environment)
        self.configuration = environment_configuration
        if desired_instances is None:
            self.desired_instances = self.configuration['cluster']['min_instances']
        else:
            self.desired_instances = desired_instances
        self.private_subnets = []
        self.public_subnets = []
        self._get_availability_zones()

    def generate_cluster(self):
        self.__validate_parameters()
        self._setup_network(
            self.configuration['vpc']['cidr'],
            self.configuration['vpc']['subnets'],
            self.configuration['vpc']['nat-gateway']['elastic-ip-allocation-id'],
        )
        self._create_log_group()
        self._add_cluster_outputs()
        self._add_cluster_parameters()
        self._add_mappings()
        self._add_metadata()
        self._add_cluster()
        return to_yaml(json.dumps(self.template.to_dict(), cls=DecimalEncoder))

    def _get_availability_zones(self):
        client = get_client_for('ec2', self.env)
        aws_azs = client.describe_availability_zones()['AvailabilityZones']
        self.availability_zones = [
            zone['ZoneName'] for zone in aws_azs
        ][:2]

    def __validate_parameters(self):
        # TODO validate CIDR
        # TODO
        return False

    # TODO: clean up
    def _setup_network(self, cidr_block, subnet_configs, eip_allocation_id):
        self._create_vpc(cidr_block)
        self._create_public_network(subnet_configs['public'])
        self._create_private_network(
            subnet_configs['private'],
            eip_allocation_id
        )
        self._create_database_subnet_group()

    def _create_vpc(self, cidr_block):
        self.vpc = VPC(
            camelcase("{self.env}Vpc".format(**locals())),
            CidrBlock=cidr_block,
            EnableDnsSupport=True,
            EnableDnsHostnames=True,
            InstanceTenancy='default',
            Tags=[
                {'Key': 'category', 'Value': 'services'},
                {'Key': 'environment', 'Value': self.env},
                {'Key': 'Name', 'Value': "{self.env}-vpc".format(**locals())}]
        )
        self.template.add_resource(self.vpc)
        self.internet_gateway = InternetGateway(
            camelcase("{self.env}Ig".format(**locals())),
            Tags=[
                {
                    'Key': 'Name',
                    'Value': "{self.env}-internet-gateway".format(**locals())
                },
                {'Key': 'environment', 'Value': self.env}
            ]
        )
        self.template.add_resource(self.internet_gateway)
        vpc_gateway_attachment = VPCGatewayAttachment(
            camelcase("{self.env}Attachment".format(**locals())),
            InternetGatewayId=Ref(self.internet_gateway),
            VpcId=Ref(self.vpc)
        )
        self.template.add_resource(vpc_gateway_attachment)
        return None

    def _create_public_network(self, subnet_configs):
        public_route_table = RouteTable(
            camelcase("{self.env}Public".format(**locals())),
            VpcId=Ref(self.vpc),
            Tags=[
                {
                    'Key': 'Name',
                    'Value': "{self.env}-public".format(**locals())
                },
                {'Key': 'environment', 'Value': self.env}
            ],
            DependsOn=self.vpc.title)
        self.template.add_resource(public_route_table)
        subnet_count = 0
        for subnet_title, subnet_config in subnet_configs.items():
            subnet_count += 1
            if subnet_count % 2 == 0:
                availability_zone = self.availability_zones[0]
            else:
                availability_zone = self.availability_zones[1]

            subnet_title = camelcase("{self.env}Public".format(**locals())) + \
                pascalcase(re.sub('[^a-zA-Z0-9*]', '', subnet_title))
            subnet_name = "{self.env}-public-{subnet_count}".format(**locals())
            subnet = Subnet(
                subnet_title,
                AvailabilityZone=availability_zone,
                CidrBlock=subnet_config['cidr'],
                VpcId=Ref(self.vpc),
                MapPublicIpOnLaunch=True,
                Tags=[
                    {'Key': 'Name', 'Value': subnet_name},
                    {'Key': 'environment', 'Value': self.env}
                ]
            )
            self.public_subnets.append(subnet)
            self.template.add_resource(subnet)
            subnet_route_table_association = SubnetRouteTableAssociation(
                camelcase("{self.env}PublicSubnet{subnet_count}Assoc".format(**locals())),
                RouteTableId=Ref(public_route_table),
                SubnetId=Ref(subnet)
            )
            self.template.add_resource(subnet_route_table_association)

        internet_gateway_route = Route(
            camelcase("{self.env}IgRoute".format(**locals())),
            DestinationCidrBlock='0.0.0.0/0',
            GatewayId=Ref(self.internet_gateway),
            RouteTableId=Ref(public_route_table)
        )
        self.template.add_resource(internet_gateway_route)
        return None

    def _create_private_network(self, subnet_configs, eip_allocation_id):
        private_route_table = RouteTable(
            camelcase("{self.env}Private".format(**locals())),
            VpcId=Ref(self.vpc),
            Tags=[
                {
                    'Key': 'Name',
                    'Value': "{self.env}-private".format(**locals())
                },
                {'Key': 'environment', 'Value': self.env}
            ]
        )
        self.template.add_resource(private_route_table)
        subnet_count = 0
        for subnet_title, subnet_config in subnet_configs.items():
            subnet_count += 1
            if subnet_count % 2 == 0:
                availability_zone = self.availability_zones[0]
            else:
                availability_zone = self.availability_zones[1]
            subnet_title = camelcase("{self.env}Private".format(**locals())) + \
                pascalcase(re.sub('[^a-zA-Z0-9*]', '', subnet_title))
            subnet_name = "{self.env}-private-{subnet_count}".format(**locals())
            subnet = Subnet(
                subnet_title,
                AvailabilityZone=availability_zone,
                CidrBlock=subnet_config['cidr'],
                VpcId=Ref(self.vpc),
                MapPublicIpOnLaunch=False,
                Tags=[
                    {'Key': 'Name', 'Value': subnet_name},
                    {'Key': 'environment', 'Value': self.env}
                ]
            )
            self.private_subnets.append(subnet)
            self.template.add_resource(subnet)
            subnet_route_table_association = SubnetRouteTableAssociation(
                camelcase("{self.env}PrivateSubnet{subnet_count}Assoc".format(
                    **locals())),
                RouteTableId=Ref(private_route_table),
                SubnetId=Ref(subnet)
            )
            self.template.add_resource(subnet_route_table_association)

        nat_gateway = NatGateway(
            camelcase("{self.env}Nat".format(**locals())),
            AllocationId=eip_allocation_id,
            SubnetId=Ref(self.public_subnets[0]),
            Tags=[
                {
                    'Key': 'Name',
                    'Value': "{self.env}-nat-gateway".format(**locals())
                },
                {'Key': 'environment', 'Value': self.env}
            ]
        )
        self.template.add_resource(nat_gateway)
        nat_gateway_route = Route(
            camelcase("{self.env}NatRoute".format(**locals())),
            DestinationCidrBlock='0.0.0.0/0',
            NatGatewayId=Ref(nat_gateway),
            RouteTableId=Ref(private_route_table)
        )
        self.template.add_resource(nat_gateway_route)
        return None

    def _create_database_subnet_group(self):
        database_subnet_group = DBSubnetGroup(
            "DBSubnetGroup",
            DBSubnetGroupName="{self.env}-subnet".format(**locals()),
            Tags=[
                {'Key': 'category', 'Value': 'services'},
                {'Key': 'environment', 'Value': self.env}
            ],
            DBSubnetGroupDescription="{self.env} subnet group".format(
                **locals()),
            SubnetIds=[Ref(subnet) for subnet in self.private_subnets]
        )
        self.template.add_resource(database_subnet_group)
        elasticache_subnet_group = ElastiCacheSubnetGroup(
            "ElasticacheSubnetGroup",
            CacheSubnetGroupName="{self.env}-subnet".format(**locals()),
            Description="{self.env} subnet group".format(**locals()),
            SubnetIds=[Ref(subnet) for subnet in self.private_subnets]
        )
        self.template.add_resource(elasticache_subnet_group)

    def _create_log_group(self):
        log_group = LogGroup(
            camelcase("{self.env}LogGroup".format(**locals())),
            LogGroupName="{self.env}-logs".format(**locals()),
            RetentionInDays=365
        )
        self.template.add_resource(log_group)
        return None

    def _create_notification_sns(self):
        return None

    def _add_instance_profile(self):
        role_name = Sub('ecs-${AWS::StackName}-${AWS::Region}')
        assume_role_policy = {
            u'Statement': [
                {
                    u'Action': [
                        u'sts:AssumeRole'
                    ],
                    u'Effect': u'Allow',
                    u'Principal': {
                        u'Service': [
                            u'ec2.amazonaws.com'
                        ]
                    }
                }
            ]
        }
        ecs_role = Role(
            'ECSRole',
            Path='/',
            ManagedPolicyArns=[
                'arn:aws:iam::aws:policy/service-role/AmazonEC2ContainerServiceforEC2Role',
                'arn:aws:iam::aws:policy/AmazonDynamoDBReadOnlyAccess',
                'arn:aws:iam::aws:policy/service-role/AmazonEC2RoleforSSM'
            ],
            RoleName=role_name,
            AssumeRolePolicyDocument=assume_role_policy
        )
        self.template.add_resource(ecs_role)
        instance_profile = InstanceProfile(
            "InstanceProfile",
            Path='/',
            Roles=[
                Ref(ecs_role)
            ]
        )
        self.template.add_resource(instance_profile)
        return instance_profile

    def _add_cluster(self):
        cluster = Cluster('Cluster', ClusterName=Ref('AWS::StackName'))
        self.template.add_resource(cluster)
        self._add_ec2_auto_scaling()
        self._add_cluster_alarms(cluster)
        return cluster

    def _add_cluster_alarms(self, cluster):
        ec2_hosts_high_cpu_alarm = Alarm(
            'Ec2HostsHighCPUAlarm',
            EvaluationPeriods=1,
            Dimensions=[
                MetricDimension(Name='AutoScalingGroupName',
                                Value=Ref(self.auto_scaling_group))
            ],
            AlarmActions=[Ref(self.notification_sns_arn)],
            AlarmDescription='Alarm if CPU too high or metric disappears \
indicating instance is down',
            Namespace='AWS/EC2',
            Period=60,
            ComparisonOperator='GreaterThanThreshold',
            Statistic='Average',
            Threshold='60',
            MetricName='CPUUtilization'
        )
        self.template.add_resource(ec2_hosts_high_cpu_alarm)
        cluster_high_cpu_alarm = Alarm(
            'ClusterHighCPUAlarm',
            EvaluationPeriods=1,
            Dimensions=[
                MetricDimension(Name='ClusterName', Value=Ref(cluster))
            ],
            AlarmActions=[
                Ref(self.notification_sns_arn)
            ],
            AlarmDescription='Alarm if CPU is too high for cluster.',
            Namespace='AWS/ECS',
            Period=300,
            ComparisonOperator='GreaterThanThreshold',
            Statistic='Average',
            Threshold='60',
            MetricName='CPUUtilization'
        )
        self.template.add_resource(cluster_high_cpu_alarm)
        cluster_high_memory_alarm = Alarm(
            'ClusterHighMemoryAlarm',
            EvaluationPeriods=1,
            Dimensions=[
                MetricDimension(Name='ClusterName', Value=Ref(cluster))
            ],
            AlarmActions=[
                Ref(self.notification_sns_arn)
            ],
            AlarmDescription='Alarm if memory is too high for cluster.',
            Namespace='AWS/ECS',
            Period=300,
            ComparisonOperator='GreaterThanThreshold',
            Statistic='Average',
            Threshold='60',
            MetricName='MemoryUtilization'
        )
        self.template.add_resource(cluster_high_memory_alarm)
        self.cluster_high_memory_reservation_autoscale_alarm = Alarm(
            'ClusterHighMemoryReservationAlarm',
            EvaluationPeriods=1,
            Dimensions=[
                MetricDimension(Name='ClusterName', Value=Ref(cluster))
            ],
            AlarmActions=[
                Ref(self.cluster_scaling_policy)
            ],
            AlarmDescription='Alarm if memory reservation is over 75% \
for cluster.',
            Namespace='AWS/ECS',
            Period=300,
            ComparisonOperator='GreaterThanThreshold',
            Statistic='Average',
            Threshold='75',
            MetricName='MemoryReservation'
        )
        self.template.add_resource(
            self.cluster_high_memory_reservation_autoscale_alarm)
        self.cluster_high_memory_reservation_user_notification_alarm = Alarm(
            'ClusterHighMemoryReservationUserNotifcationAlarm',
            EvaluationPeriods=3,
            Dimensions=[
                MetricDimension(Name='ClusterName', Value=Ref(cluster))
            ],
            AlarmActions=[
                Ref(self.notification_sns_arn)
            ],
            OKActions=[
                Ref(self.notification_sns_arn)
            ],
            AlarmDescription='Alarm if memory reservation is over 75% \
for cluster for 15 minutes.',
            Namespace='AWS/ECS',
            Period=300,
            ComparisonOperator='GreaterThanThreshold',
            Statistic='Average',
            Threshold='75',
            MetricName='MemoryReservation'
        )
        self.template.add_resource(
            self.cluster_high_memory_reservation_user_notification_alarm)

    def _add_ec2_auto_scaling(self):
        instance_profile = self._add_instance_profile()
        self.sg_alb = SecurityGroup(
            "SecurityGroupAlb",
            VpcId=Ref(self.vpc),
            GroupDescription=Sub("${AWS::StackName}-alb")
        )
        self.template.add_resource(self.sg_alb)
        sg_hosts = SecurityGroup(
            "SecurityGroupEc2Hosts",
            SecurityGroupIngress=[
                {
                    'SourceSecurityGroupId': Ref(self.sg_alb),
                    'IpProtocol': -1
                }
            ],
            VpcId=Ref(self.vpc),
            GroupDescription=Sub("${AWS::StackName}-hosts")
        )
        self.template.add_resource(sg_hosts)
        database_security_group = SecurityGroup(
            "SecurityGroupDatabases",
            SecurityGroupIngress=[
                {
                    'SourceSecurityGroupId': Ref(sg_hosts),
                    'IpProtocol': -1
                }
            ],
            VpcId=Ref(self.vpc),
            GroupDescription=Sub("${AWS::StackName}-databases")
        )
        self.template.add_resource(database_security_group)
        user_data = Base64(Sub('\n'.join([
            "#!/bin/bash",
            "yum update -y",
            "yum install -y aws-cfn-bootstrap",
            "/opt/aws/bin/cfn-init -v --region ${AWS::Region} --stack ${AWS::StackName} --resource LaunchConfiguration",
            "/opt/aws/bin/cfn-signal -e $? --region ${AWS::Region} --stack ${AWS::StackName} --resource AutoScalingGroup",
            "yum install -y https://s3.amazonaws.com/ec2-downloads-windows/SSMAgent/latest/linux_amd64/amazon-ssm-agent.rpm",
            "systemctl enable amazon-ssm-agent",
            "systemctl start amazon-ssm-agent",
            ""])))
        lc_metadata = cloudformation.Init({
            "config": cloudformation.InitConfig(
                files=cloudformation.InitFiles({
                    "/etc/cfn/cfn-hup.conf": cloudformation.InitFile(
                        content=Sub(
                            '\n'.join([
                                    '[main]',
                                    'stack=${AWS::StackId}',
                                    'region=${AWS::Region}',
                                    ''
                                ])
                            ),
                        mode='256',  # TODO: Why 256
                        owner="root",
                        group="root"
                    ),
                    "/etc/cfn/hooks.d/cfn-auto-reloader.conf": cloudformation.InitFile(
                        content=Sub(
                            '\n'.join([
                                '[cfn-auto-reloader-hook]',
                                'triggers=post.update',
                                'path=Resources.ContainerInstances.Metadata.AWS::CloudFormation::Init',
                                'action=/opt/aws/bin/cfn-init -v --region ${AWS::Region} --stack ${AWS::StackName} --resource LaunchConfiguration',
                                ''
                            ])
                        ),
                    )
                }),
                services={
                    "sysvinit": cloudformation.InitServices({
                        "cfn-hup": cloudformation.InitService(
                            enabled=True,
                            ensureRunning=True,
                            files=['/etc/cfn/cfn-hup.conf',
                                   '/etc/cfn/hooks.d/cfn-auto-reloader.conf']
                        )
                    })
                },
                commands={
                    '01_add_instance_to_cluster': {
                        'command': Sub(
                            'echo "ECS_CLUSTER=${Cluster}\nECS_RESERVED_MEMORY=256" > /etc/ecs/ecs.config'
                        )
                    }
                }
            )
        })
        launch_configuration = LaunchConfiguration(
            'LaunchConfiguration',
            UserData=user_data,
            IamInstanceProfile=Ref(instance_profile),
            SecurityGroups=[Ref(sg_hosts)],
            InstanceType=Ref('InstanceType'),
            ImageId=FindInMap("AWSRegionToAMI", Ref("AWS::Region"), "AMI"),
            Metadata=lc_metadata,
            KeyName=Ref(self.key_pair)
        )
        self.template.add_resource(launch_configuration)
        # , PauseTime='PT15M', WaitOnResourceSignals=True, MaxBatchSize=1, MinInstancesInService=1)
        up = AutoScalingRollingUpdate('AutoScalingRollingUpdate')
        # TODO: clean up
        subnets = list(self.private_subnets)
        self.auto_scaling_group = AutoScalingGroup(
            "AutoScalingGroup",
            UpdatePolicy=up,
            DesiredCapacity=self.desired_instances,
            Tags=[
                {
                    'PropagateAtLaunch': True,
                    'Value': Sub('${AWS::StackName} - ECS Host'),
                    'Key': 'Name'
                }
            ],
            MinSize=Ref('MinSize'),
            MaxSize=Ref('MaxSize'),
            VPCZoneIdentifier=[Ref(subnets.pop()), Ref(subnets.pop())],
            LaunchConfigurationName=Ref(launch_configuration),
            CreationPolicy=CreationPolicy(
                ResourceSignal=ResourceSignal(Timeout='PT15M')
            )
        )
        self.template.add_resource(self.auto_scaling_group)
        self.cluster_scaling_policy = ScalingPolicy(
            'AutoScalingPolicy',
            AdjustmentType='ChangeInCapacity',
            AutoScalingGroupName=Ref(self.auto_scaling_group),
            Cooldown=300,
            PolicyType='SimpleScaling',
            ScalingAdjustment=1
        )
        self.template.add_resource(self.cluster_scaling_policy)

    def _add_cluster_parameters(self):
        self.template.add_parameter(Parameter(
            "Environment",
            Description='',
            Type="String",
            Default="")
        )
        self.key_pair = Parameter(
            "KeyPair", Description='', Type="AWS::EC2::KeyPair::KeyName", Default="")
        self.template.add_parameter(self.key_pair)
        self.template.add_parameter(Parameter(
            "MinSize", Description='', Type="Number", Default=str(self.configuration['cluster']['min_instances'])))
        self.template.add_parameter(Parameter(
            "MaxSize", Description='', Type="Number", Default=str(self.configuration['cluster']['max_instances'])))
        self.notification_sns_arn = Parameter("NotificationSnsArn",
                                              Description='',
                                              Type="String",
                                              Default=self.notifications_arn)
        self.template.add_parameter(self.notification_sns_arn)
        self.template.add_parameter(Parameter(
            "InstanceType", Description='', Type="String", Default=self.configuration['cluster']['instance_type']))

    def _add_mappings(self):
        # Pick from https://docs.aws.amazon.com/AmazonECS/latest/developerguide/al2ami.html
        ssm_client = get_client_for('ssm', self.env)
        ami_response = ssm_client.get_parameter(
            Name='/aws/service/ecs/optimized-ami/amazon-linux-2/recommended')
        ami_id = json.loads(ami_response['Parameter']['Value'])['image_id']
        region = get_region_for_environment(self.env)
        self.template.add_mapping('AWSRegionToAMI', {
            region: {"AMI": ami_id}
        })

    def _add_cluster_outputs(self):
        self._add_stack_outputs()
        metadata = {
            'env': self.env,
            'min_instances': str(self.configuration['cluster']['min_instances']),
            'max_instances': str(self.configuration['cluster']['max_instances']),
            'instance_type': self.configuration['cluster']['instance_type'],
            'key_name': self.configuration['cluster']['key_name'],
            'cloudlift_version': VERSION
        }
        self.template.add_output(Output(
            "CloudliftOptions",
            Description="Options used with cloudlift when building this cluster",
            Value=json.dumps(metadata))
        )
        self.template.add_output(Output(
            "VPC",
            Description="VPC in which environment is setup",
            Value=Ref(self.vpc))
        )
        private_subnets = list(self.private_subnets)
        self.template.add_output(Output(
            "PrivateSubnet1",
            Description="ID of the 1st subnet",
            Value=Ref(private_subnets.pop()))
        )
        self.template.add_output(Output(
            "PrivateSubnet2",
            Description="ID of the 2nd subnet",
            Value=Ref(private_subnets.pop()))
        )
        public_subnets = list(self.public_subnets)
        self.template.add_output(Output(
            "PublicSubnet1",
            Description="ID of the 1st subnet",
            Value=Ref(public_subnets.pop()))
        )
        self.template.add_output(Output(
            "PublicSubnet2",
            Description="ID of the 2nd subnet",
            Value=Ref(public_subnets.pop()))
        )
        self.template.add_output(Output(
            "AutoScalingGroup",
            Description="AutoScaling group for ECS container instances",
            Value=Ref('AutoScalingGroup'))
        )
        self.template.add_output(Output(
            "SecurityGroupAlb",
            Description="Security group ID for ALB",
            Value=Ref('SecurityGroupAlb'))
        )
        self.template.add_output(Output(
            "MinInstances",
            Description="Minimum instances in cluster",
            Value=str(self.configuration['cluster']['min_instances']))
        )
        self.template.add_output(Output(
            "MaxInstances",
            Description="Maximum instances in cluster",
            Value=str(self.configuration['cluster']['max_instances']))
        )
        self.template.add_output(Output(
            "InstanceType",
            Description="EC2 instance type",
            Value=str(self.configuration['cluster']['instance_type']))
        )
        self.template.add_output(Output(
            "KeyName",
            Description="Key Pair name for accessing the instances",
            Value=str(self.configuration['cluster']['key_name']))
        )

    def _add_metadata(self):
        self.template.add_metadata({
            'AWS::CloudFormation::Interface': {
                'ParameterGroups': [
                    {
                        'Label': {
                            'default': 'Cluster Configuration'
                        },
                        'Parameters': [
                            'KeyPair',
                            'Environment',
                            'MinSize',
                            'MaxSize',
                            'InstanceType',
                            'VPC',
                            'Subnet1',
                            'Subnet2',
                            'NotificationSnsArn'
                        ]
                    },
                ],
                'ParameterLabels': {
                    'Environment': {
                        'default': 'Enter the environment e.g. dev or staging or sandbox or production'
                    },
                    'InstanceType': {
                        'default': 'Type of instance'
                    },
                    'KeyPair': {
                        'default': 'Select the key with which you want to login to the ec2 instances'},
                    'MaxSize': {
                        'default': 'Max. no. of instances in cluster'
                    },
                    'MinSize': {
                        'default': 'Min. no. of instances in cluster'
                    },
                    'NotificationSnsArn': {
                        'default': 'The SNS topic to which notifactions has to be triggered'
                    },
                    'Subnet1': {
                        'default': 'Enter the ID of the 1st subnet'
                    },
                    'Subnet2': {
                        'default': 'Enter the ID of the 2nd subnet'
                    },
                    'VPC': {
                        'default': 'Enter the VPC in which you want the environment to be setup'
                    },
                }
            }
        })



import functools

import boto3
import click
from botocore.exceptions import ClientError

from cloudlift.config import highlight_production
from cloudlift.deployment.configs import deduce_name
from cloudlift.deployment import EnvironmentCreator, editor
from cloudlift.config.logging import log_err
from cloudlift.deployment.service_creator import ServiceCreator
from cloudlift.deployment.service_information_fetcher import ServiceInformationFetcher
from cloudlift.deployment.service_updater import ServiceUpdater
from cloudlift.deployment.task_definition_creator import TaskDefinitionCreator
from cloudlift.session import SessionCreator
from cloudlift.version import VERSION
from cloudlift.exceptions import UnrecoverableException


def _require_environment(func):
    @click.option('--environment', '-e', prompt='environment',
                  help='environment')
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if kwargs['environment'] == 'production':
            highlight_production()
        return func(*args, **kwargs)

    return wrapper


def _require_name(func):
    @click.option('--name', help='Your service name, give the name of \
repo')
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if kwargs['name'] is None:
            kwargs['name'] = deduce_name(None)
        return func(*args, **kwargs)

    return wrapper


class CommandWrapper(click.Group):
    def __call__(self, *args, **kwargs):
        try:
            return self.main(*args, **kwargs)
        except UnrecoverableException as e:
            log_err(e.value)
            exit(1)


@click.group(cls=CommandWrapper)
@click.version_option(version=VERSION, prog_name="cloudlift")
def cli():
    """
        Cloudlift is built by Simpl developers to make it easier to launch \
        dockerized services in AWS ECS.
    """
    try:
        boto3.client('cloudformation')
    except ClientError:
        log_err("Could not connect to AWS!")
        log_err("Ensure AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY & \
AWS_DEFAULT_REGION env vars are set OR run 'aws configure'")
        exit(1)


@cli.command(help="Create a new service. This can contain multiple \
ECS services")
@_require_environment
@_require_name
def create_service(name, environment):
    ServiceCreator(name, environment).create()


@cli.command(help="Update existing service.")
@_require_environment
@_require_name
def update_service(name, environment):
    ServiceCreator(name, environment).update()


@cli.command(help="Create a new environment")
@click.option('--environment', '-e', prompt='environment',
              help='environment')
def create_environment(environment):
    EnvironmentCreator(environment).run()


@cli.command(help="Update environment")
@_require_environment
@click.option('--update_ecs_agents',
              is_flag=True,
              help='Update ECS container agents')
def update_environment(environment, update_ecs_agents):
    EnvironmentCreator(environment).run_update(update_ecs_agents)


@cli.command(help="Command used to create or update the configuration \
in parameter store")
@_require_name
@_require_environment
def edit_config(name, environment):
    editor.edit_config(name, environment)


@cli.command()
@_require_environment
@_require_name
@click.option('--version', default=None,
              help='local image version tag')
@click.option("--build-arg", type=(str, str), multiple=True, help="These args are passed to docker build command "
                                                                  "as --build-args. Supports multiple.\
                                                                   Please leave space between name and value" )
def deploy_service(name, environment, version, build_arg):
    ServiceUpdater(name, environment, None, version, dict(build_arg)).run()


@cli.command()
@_require_environment
@_require_name
@click.option('--version', default=None,
              help='local image version tag')
@click.option("--build-arg", type=(str, str), multiple=True, help="These args are passed to docker build command "
                                                                  "as --build-args. Supports multiple.\
                                                                   Please leave space between name and value" )
def create_task_definition(name, environment, version, build_arg):
    TaskDefinitionCreator(name, environment, version, dict(build_arg)).create()


@cli.command()
@_require_environment
@_require_name
@click.option('--version', default=None,
              help='local image version tag')
@click.option("--build-arg", type=(str, str), multiple=True, help="These args are passed to docker build command "
                                                                  "as --build-args. Supports multiple.\
                                                                   Please leave space between name and value" )
def update_task_definition(name, environment, version, build_arg):
    TaskDefinitionCreator(name, environment, version, dict(build_arg)).update()


@cli.command()
@click.option('--local_tag', help='Commit sha for image to be uploaded')
@click.option('--additional_tags', default=[], multiple=True,
              help='Additional tags for the image apart from commit SHA')
@_require_name
def upload_to_ecr(name, local_tag, additional_tags):
    ServiceUpdater(name, '', '', local_tag).upload_image(additional_tags)


@cli.command(help="Get commit information of currently deployed code \
from commit hash")
@_require_environment
@_require_name
@click.option('--short', '-s', is_flag=True,
              help='Pass this when you just need the version tag')
def get_version(name, environment, short):
    ServiceInformationFetcher(name, environment).get_version(short)


@cli.command(help="Start SSH session in instance running a current \
service task")
@_require_environment
@_require_name
@click.option('--mfa', help='MFA code', prompt='MFA Code')
def start_session(name, environment, mfa):
    SessionCreator(name, environment).start_session(mfa)


if __name__ == '__main__':
    cli()



'''
Create or update ECS service and related dependencies
using CloudFormation templates
'''

from time import sleep

from botocore.exceptions import ClientError
from cloudlift.exceptions import UnrecoverableException

from cloudlift.config import get_client_for
from cloudlift.config import ServiceConfiguration
from cloudlift.config import get_cluster_name, get_service_stack_name
from cloudlift.deployment.changesets import create_change_set
from cloudlift.config.logging import log, log_bold, log_err
from cloudlift.deployment.progress import get_stack_events, print_new_events
from cloudlift.deployment.service_template_generator import ServiceTemplateGenerator


class ServiceCreator(object):
    '''
        Create and execute CloudFormation template or changeset for existing
        CloudFormation template for ECS service and related dependencies
    '''

    def __init__(self, name, environment):
        self.name = name
        self.environment = environment
        self.stack_name = get_service_stack_name(environment, name)
        self.client = get_client_for('cloudformation', self.environment)
        self.s3client = get_client_for('s3', self.environment)
        self.bucket_name = 'cloudlift-service-template'
        self.environment_stack = self._get_environment_stack()
        self.existing_events = get_stack_events(self.client, self.stack_name)
        self.service_configuration = ServiceConfiguration(
            self.name,
            self.environment
        )

    def delete_template(self, key=None):
        '''
            Delete CloudFormation Template stored in S3 bucket
        '''
        if key:
            try:
                self.s3client.delete_object(
                    Bucket=self.bucket_name,
                    Key=key,
                )
            except ClientError as boto_client_error:
                print(f"Error deleting s3 key: {key}")
                raise boto_client_error

    def create(self):
        '''
            Create and execute CloudFormation template for ECS service
            and related dependencies
        '''
        log_bold("Initiating service creation")
        self.service_configuration.edit_config()

        template_generator = ServiceTemplateGenerator(
            self.service_configuration,
            self.environment_stack
        )
        service_template_body, template_source, key = template_generator.generate_service()

        try:
            if template_source == 'TemplateBody':
                self.client.create_stack(
                    StackName=self.stack_name,
                    TemplateBody=service_template_body,
                    Parameters=[{
                        'ParameterKey': 'Environment',
                        'ParameterValue': self.environment,
                    }],
                    OnFailure='DO_NOTHING',
                    Capabilities=['CAPABILITY_NAMED_IAM'],
                )
            elif template_source == 'TemplateURL':
                self.client.create_stack(
                    StackName=self.stack_name,
                    TemplateURL=service_template_body,
                    Parameters=[{
                        'ParameterKey': 'Environment',
                        'ParameterValue': self.environment,
                    }],
                    OnFailure='DO_NOTHING',
                    Capabilities=['CAPABILITY_NAMED_IAM'],
                )
            log_bold("Submitted to cloudformation. Checking progress...")
            self.delete_template(key)
            self._print_progress()
        except ClientError as boto_client_error:
            self.delete_template(key)
            error_code = boto_client_error.response['Error']['Code']
            if error_code == 'AlreadyExistsException':
                raise UnrecoverableException("Stack " + self.stack_name + " already exists.")
            else:
                raise boto_client_error

    def update(self):
        '''
            Create and execute changeset for existing CloudFormation template
            for ECS service and related dependencies
        '''

        log_bold("Starting to update service")
        self.service_configuration.edit_config()
        try:
            template_generator = ServiceTemplateGenerator(
                self.service_configuration,
                self.environment_stack
            )
            service_template_body, template_source, key = template_generator.generate_service()
            change_set = create_change_set(
                self.client,
                service_template_body,
                template_source,
                self.stack_name,
                "",
                self.environment
            )
            if change_set is None:
                self.delete_template(key)
                return
            self.service_configuration.update_cloudlift_version()
            log_bold("Executing changeset. Checking progress...")
            self.client.execute_change_set(
                ChangeSetName=change_set['ChangeSetId']
            )
            self.delete_template(key)
            self._print_progress()
        except ClientError as exc:
            self.delete_template(key)
            if "No updates are to be performed." in str(exc):
                log_err("No updates are to be performed")
            else:
                raise exc

    def _get_environment_stack(self):
        try:
            log("Looking for " + self.environment + " cluster.")
            environment_stack = self.client.describe_stacks(
                StackName=get_cluster_name(self.environment)
            )['Stacks'][0]
            log_bold(self.environment+" stack found. Using stack with ID: " +
                     environment_stack['StackId'])
        except ClientError:
            raise UnrecoverableException(self.environment + " cluster not found. Create the environment \
cluster using `create_environment` command.")
        return environment_stack

    def _print_progress(self):
        while True:
            response = self.client.describe_stacks(StackName=self.stack_name)
            if "IN_PROGRESS" not in response['Stacks'][0]['StackStatus']:
                break
            all_events = get_stack_events(self.client, self.stack_name)
            print_new_events(all_events, self.existing_events)
            self.existing_events = all_events
            sleep(5)
        final_status = response['Stacks'][0]['StackStatus']
        if "FAIL" in final_status:
            log_err("Finished with status: %s" % (final_status))
        else:
            log_bold("Finished with status: %s" % (final_status))



"""
This module handles global cloudlift configuration that is custom to
the organization using cloudlift
"""
import ipaddress
import json

import boto3
import dictdiffer
from botocore.exceptions import ClientError
from click import confirm, edit, prompt
from jsonschema import validate
from jsonschema.exceptions import ValidationError

from cloudlift.exceptions import UnrecoverableException
from cloudlift.config import DecimalEncoder
from cloudlift.config import print_json_changes
# import config.mfa as mfa
from cloudlift.config.logging import log_bold, log_err, log_warning

ENVIRONMENT_CONFIGURATION_TABLE = 'environment_configurations'

class EnvironmentConfiguration(object):
    '''
        Handles configuration in DynamoDB for cloudlift
    '''

    def __init__(self, environment=None):
        self.environment = environment

        session = boto3.session.Session()
        self.dynamodb = session.resource('dynamodb')
        self.table = self._get_table()

    def get_config(self):
        '''
            Get configuration from DynamoDB
        '''

        try:
            configuration_response = self.table.get_item(
                Key={
                    'environment': self.environment
                },
                ConsistentRead=True,
                AttributesToGet=[
                    'configuration'
                ]
            )
            return configuration_response['Item']['configuration']
        except ClientError:
            raise UnrecoverableException("Unable to fetch environment configuration from DynamoDB.")
        except KeyError:
            raise UnrecoverableException("Environment configuration not found. Does this environment exist?")

    def update_config(self):
        if not self._env_config_exists():
            self._create_config()
        self._edit_config()

    def get_all_environments(self):
        response = self.table.scan(
            TableName=ENVIRONMENT_CONFIGURATION_TABLE,
            AttributesToGet=[
                'environment',
            ],
            Limit=10,
            Select='SPECIFIC_ATTRIBUTES',
            ScanFilter={},
            ConsistentRead=True
        )
        return [env['environment'] for env in response['Items']]

    def _get_table(self):
        dynamodb_client = boto3.session.Session().client('dynamodb')
        table_names = dynamodb_client.list_tables()['TableNames']
        if ENVIRONMENT_CONFIGURATION_TABLE not in table_names:
            log_warning("Could not find configuration table, creating one..")
            self._create_configuration_table()
        return self.dynamodb.Table(ENVIRONMENT_CONFIGURATION_TABLE)

    def _create_configuration_table(self):
        self.dynamodb.create_table(
            TableName=ENVIRONMENT_CONFIGURATION_TABLE,
            KeySchema=[
                {
                    'AttributeName': 'environment',
                    'KeyType': 'HASH'
                }
            ],
            AttributeDefinitions=[
                {
                    'AttributeName': 'environment',
                    'AttributeType': 'S'
                }
            ],
            BillingMode='PAY_PER_REQUEST'
        )
        log_bold("Configuration table created!")

    def _env_config_exists(self):
        response = self.table.get_item(
            Key={
                'environment': self.environment,
            }
        )
        return response.get('Item') is not None

    def _create_config(self):
        log_warning(
            "\nConfiguration for this environment was not found in DynamoDB.\
            \nInitiating prompts for setting up configuration.\
            \nIf this environment was previously configured, please ensure\
            \ndefault region is the same as previously used. Otherwise, use\
            \nthe same configuration.\n"
        )
        region = prompt("AWS region for environment", default='ap-south-1')
        vpc_cidr = ipaddress.IPv4Network(prompt("VPC CIDR", default='10.10.10.10/16'))
        nat_eip = prompt("Allocation ID Elastic IP for NAT")
        public_subnet_1_cidr = prompt(
            "Public Subnet 1 CIDR", default=list(vpc_cidr.subnets(new_prefix=22))[0])
        public_subnet_2_cidr = prompt(
            "Public Subnet 2 CIDR", default=list(vpc_cidr.subnets(new_prefix=22))[1])
        private_subnet_1_cidr = prompt(
            "Private Subnet 1 CIDR", default=list(vpc_cidr.subnets(new_prefix=22))[2])
        private_subnet_2_cidr = prompt(
            "Private Subnet 2 CIDR", default=list(vpc_cidr.subnets(new_prefix=22))[3])
        cluster_min_instances = prompt("Min instances in cluster", default=1)
        cluster_max_instances = prompt("Max instances in cluster", default=5)
        cluster_instance_type = prompt("Instance type", default='m5.xlarge')
        key_name = prompt("SSH key name")
        notifications_arn = prompt("Notification SNS ARN")
        ssl_certificate_arn = prompt("SSL certificate ARN")
        environment_configuration = {
            self.environment: {
                "region": region,
                "vpc": {
                    "cidr": str(vpc_cidr),
                    "nat-gateway": {
                        "elastic-ip-allocation-id": nat_eip
                    },
                    "subnets": {
                        "public": {
                            "subnet-1": {
                                "cidr": str(public_subnet_1_cidr)
                            },
                            "subnet-2": {
                                "cidr": str(public_subnet_2_cidr)
                            }
                        },
                        "private": {
                            "subnet-1": {
                                "cidr": str(private_subnet_1_cidr)
                            },
                            "subnet-2": {
                                "cidr": str(private_subnet_2_cidr)
                            }
                        }
                    }
                },
                "cluster": {
                    "min_instances": cluster_min_instances,
                    "max_instances": cluster_max_instances,
                    "instance_type": cluster_instance_type,
                    "key_name": key_name
                },
                "environment": {
                    "notifications_arn": notifications_arn,
                    "ssl_certificate_arn": ssl_certificate_arn
                }
            }
        }
        self._set_config(environment_configuration)
        pass

    def _edit_config(self):
        '''
            Open editor to update configuration
        '''

        try:
            current_configuration = self.get_config()

            updated_configuration = edit(
                json.dumps(
                    current_configuration,
                    indent=4,
                    sort_keys=True,
                    cls=DecimalEncoder
                )
            )

            if updated_configuration is None:
                log_warning("No changes made.")
            else:
                updated_configuration = json.loads(updated_configuration)
                differences = list(dictdiffer.diff(
                    current_configuration,
                    updated_configuration
                ))
                if not differences:
                    log_warning("No changes made.")
                else:
                    print_json_changes(differences)
                    if confirm('Do you want update the config?'):
                        self._set_config(updated_configuration)
                    else:
                        log_warning("Changes aborted.")
        except ClientError:
            raise UnrecoverableException("Unable to fetch environment configuration from DynamoDB.")

    def _set_config(self, config):
        '''
            Set configuration in DynamoDB
        '''
        self._validate_changes(config)
        try:
            configuration_response = self.table.update_item(
                TableName=ENVIRONMENT_CONFIGURATION_TABLE,
                Key={
                    'environment': self.environment
                },
                UpdateExpression='SET configuration = :configuration',
                ExpressionAttributeValues={
                    ':configuration': config
                },
                ReturnValues="UPDATED_NEW"
            )
            return configuration_response
        except ClientError:
            raise UnrecoverableException("Unable to store environment configuration in DynamoDB.")
        pass

    def _validate_changes(self, configuration):
        log_bold("\nValidating schema..")
        # TODO: add cidr etc validation
        schema = {
            # "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "configuration",
            "type": "object",
            "properties": {
                self.environment: {
                    "type": "object",
                    "properties": {
                        "cluster": {
                            "type": "object",
                            "properties": {
                                "min_instances": {"type": "integer"},
                                "max_instances": {"type": "integer"},
                                "instance_type": {"type": "string"},
                                "key_name": {"type": "string"},
                            },
                            "required": [
                                "min_instances",
                                "max_instances",
                                "instance_type",
                                "key_name"
                            ]
                        },
                        "environment": {
                            "type": "object",
                            "properties": {
                                "notifications_arn": {"type": "string"},
                                "ssl_certificate_arn": {"type": "string"}
                            },
                            "required": [
                                "notifications_arn",
                                "ssl_certificate_arn"
                            ]
                        },
                        "region": {"type": "string"},
                        "vpc": {
                            "type": "object",
                            "properties": {
                                "cidr": {
                                    "type": "string"
                                },
                                "nat-gateway": {
                                    "type": "object",
                                    "properties": {
                                        "elastic-ip-allocation-id": {
                                            "type": "string"
                                        }
                                    },
                                    "required": [
                                        "elastic-ip-allocation-id"
                                    ]
                                },
                                "subnets": {
                                    "type": "object",
                                    "properties": {
                                        "private": {
                                            "type": "object",
                                            "properties": {
                                                "subnet-1": {
                                                    "type": "object",
                                                    "properties": {
                                                        "cidr": {
                                                            "type": "string"
                                                        }
                                                    },
                                                    "required": [
                                                        "cidr"
                                                    ]
                                                },
                                                "subnet-2": {
                                                    "type": "object",
                                                    "properties": {
                                                        "cidr": {
                                                            "type": "string"
                                                        }
                                                    },
                                                    "required": [
                                                        "cidr"
                                                    ]
                                                }
                                            },
                                            "required": [
                                                "subnet-1",
                                                "subnet-2"
                                            ]
                                        },
                                        "public": {
                                            "type": "object",
                                            "properties": {
                                                "subnet-1": {
                                                    "type": "object",
                                                    "properties": {
                                                        "cidr": {
                                                            "type": "string"
                                                        }
                                                    },
                                                    "required": [
                                                        "cidr"
                                                    ]
                                                },
                                                "subnet-2": {
                                                    "type": "object",
                                                    "properties": {
                                                        "cidr": {
                                                            "type": "string"
                                                        }
                                                    },
                                                    "required": [
                                                        "cidr"
                                                    ]
                                                }
                                            },
                                            "required": [
                                                "subnet-1",
                                                "subnet-2"
                                            ]
                                        }
                                    },
                                    "required": [
                                        "private",
                                        "public"
                                    ]
                                }
                            },
                            "required": [
                                "cidr",
                                "nat-gateway",
                                "subnets"
                            ]
                        }

                    },
                    "required": [
                        "cluster",
                        "environment",
                        "region",
                        "vpc"
                    ]
                }
            },
            "required": [self.environment]
        }
        try:
            validate(configuration, schema)
        except ValidationError as validation_error:
            error_path = str(".".join(list(validation_error.relative_path)))
            if error_path:
                raise UnrecoverableException(validation_error.message + " in " + error_path)
            else:
                raise UnrecoverableException(validation_error.message)
        log_bold("Schema valid!")



from .account import *
from .diff import *
from .decimal_encoder import *
from .environment_configuration import *
from .region import *
from .parameter_store import *
from .banner import *
from .mfa import *
from .service_configuration import *
from .stack import *


'''
    Handles decimal conversion to json
    https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/GettingStarted.Python.03.html
'''

import decimal
import json


class DecimalEncoder(json.JSONEncoder):
    '''
      Handles decimal conversion to json
    '''

    def default(self, o):
        '''
            Extend default from json and replace decimals with
            float or int
        '''

        if isinstance(o, decimal.Decimal):
            if abs(o) % 1 > 0:
                return float(o)
            else:
                return int(o)
        return super(DecimalEncoder, self).default(o)



import os

from stringcase import pascalcase
from troposphere.ecs import (ContainerDefinition,
                             Environment,
                             LogConfiguration,
                             TaskDefinition)

from cloudlift.config.logging import log_bold, log_intent, log_warning
from cloudlift.deployment import EcrClient, UnrecoverableException, EcsClient, DeployAction, EcsTaskDefinition
from cloudlift.deployment.deployer import build_config, print_task_diff


def _complete_image_url(ecr_client: EcrClient):
    return ecr_client.ecr_image_uri + ':' + ecr_client.version


class TaskDefinitionCreator:
    def __init__(self, name, environment, version, build_args, region='ap-south-1'):
        self.name = name
        self.environment = environment
        self.build_args = build_args
        self.region = region
        self.version = version
        self.env_sample_file = './env.sample'
        self.cluster_name = f'cluster-{self.environment}'
        self.name_with_env = f"{pascalcase(self.name)}{pascalcase(self.environment)}"

    def create(self):
        log_warning("Create task definition to {self.region}".format(**locals()))
        if not os.path.exists(self.env_sample_file):
            raise UnrecoverableException('env.sample not found. Exiting.')
        ecr_client = EcrClient(self.name, self.region, self.build_args)
        ecr_client.set_version(self.version)
        log_intent("name: " + self.name + " | environment: " +
                   self.environment + " | version: " + str(ecr_client.version))
        log_bold("Checking image in ECR")
        ecr_client.build_and_upload_image()
        log_bold("Creating task definition\n")
        env_config = build_config(self.environment, self.name, self.env_sample_file)
        container_definition_arguments = {
            "environment": [
                {
                    "name": k,
                    "value": v
                } for (k, v) in env_config
            ],
            "name": pascalcase(self.name) + "Container",
            "image": _complete_image_url(ecr_client),
            "essential": True,
            "logConfiguration": self._gen_log_config(pascalcase(self.name)),
            "memoryReservation": 1024
        }

        ecs_client = EcsClient(region=self.region)
        ecs_client.register_task_definition(self._task_defn_family(), [container_definition_arguments], [], None)
        log_bold("Task definition successfully created\n")

    def update(self):
        log_warning("Update task definition to {self.region}".format(**locals()))
        if not os.path.exists(self.env_sample_file):
            raise UnrecoverableException('env.sample not found. Exiting.')
        ecr_client = EcrClient(self.name, self.region, self.build_args)
        ecr_client.set_version(self.version)
        log_intent("name: " + self.name + " | environment: " +
                   self.environment + " | version: " + str(ecr_client.version))
        log_bold("Checking image in ECR")
        ecr_client.build_and_upload_image()
        log_bold("Updating task definition\n")
        env_config = build_config(self.environment, self.name, self.env_sample_file)
        ecs_client = EcsClient(region=self.region)
        deployment = DeployAction(ecs_client, self.cluster_name, None)
        task_defn = self._apply_changes_over_current_task_defn(env_config, ecs_client, ecr_client, deployment)
        deployment.update_task_definition(task_defn)
        log_bold("Task definition successfully updated\n")

    def _current_task_defn(self, ecs_client: EcsClient, deployment: DeployAction):
        task_defn_arn = ecs_client.list_task_definitions(self._task_defn_family())[0]
        return deployment.get_task_definition(task_defn_arn)

    def _apply_changes_over_current_task_defn(self, env_config, ecs_client: EcsClient, ecr_client: EcrClient,
                                              deployment: DeployAction):
        current_task_defn = self._current_task_defn(ecs_client, deployment)
        container_name = current_task_defn['containerDefinitions'][0]['name']
        current_task_defn.set_images(
            ecr_client.version,
            **{container_name: _complete_image_url(ecr_client)}
        )
        for container in current_task_defn.containers:
            current_task_defn.apply_container_environment(container, env_config)
        print_task_diff(self.name, current_task_defn.diff, 'white')
        return current_task_defn

    def _task_defn_family(self):
        return f"{self.name_with_env}Family"

    def _gen_log_config(self, stream_prefix):
        return {
            'logDriver': 'awslogs',
            'options': {
                'awslogs-stream-prefix': stream_prefix,
                'awslogs-group': '-'.join([self.environment, 'logs']),
                'awslogs-region': self.region
            }
        }



from terminaltables import SingleTable

from cloudlift.config.logging import log_bold


def print_parameter_changes(differences):
    changes_to_show = [["Type", "Config", "Old val", "New val"]]
    for difference in differences:
        if difference[0] == 'change':
            changes_to_show.append([
                'change',
                difference[1],
                difference[2][0],
                difference[2][1]]
            )
        if difference[0] == 'add':
            difference[2].sort(key=lambda x: x[0])
            for added_item in difference[2]:
                changes_to_show.append([
                    'add',
                    added_item[0],
                    '',
                    added_item[1]])
        if difference[0] == 'remove':
            difference[2].sort(key=lambda x: x[0])
            for removed_item in difference[2]:
                changes_to_show.append(['remove', removed_item[0],
                                        removed_item[1], ''])
    log_bold("Modifications to config:")
    print(SingleTable(changes_to_show).table)


def print_json_changes(differences):
    changes_to_show = [["Type", "Config", "Old val", "New val"]]
    for difference in differences:
        if difference[0] == 'change':
            changes_to_show.append([
                'change',
                difference[1],
                difference[2][0],
                difference[2][1]
            ])
        if difference[0] == 'add':
            difference[2].sort(key=lambda x: x[0])
            for added_item in difference[2]:
                changes_to_show.append([
                    'add',
                    difference[1],
                    '',
                    str(added_item[0])+" : "+str(added_item[1])
                ])
        if difference[0] == 'remove':
            difference[2].sort(key=lambda x: x[0])
            for removed_item in difference[2]:
                changes_to_show.append([
                    'remove',
                    str(difference[1])+"."+str(removed_item[0]),
                    removed_item[1],
                    ''
                ])
    log_bold("Modifications to config:")
    print(SingleTable(changes_to_show).table)



import os

import stringcase

from cloudlift.config.logging import log_bold


def deduce_name(name):
    if name is None:
        name = os.path.basename(os.getcwd())
        log_bold("Assuming the service name to be: " + name)
    return stringcase.spinalcase(name)



"""
This is built from the package ecs_deploy.ecs
https://github.com/fabfuel/ecs-deploy/blob/develop/ecs_deploy/ecs.py

The package defaults to copying old environment config, applying
value changes onto it and merging new configs. This means, environment
variables cannot be deleted.

To update the package:
Copy the new contents, but ensure that apply_container_environment is retained
or updated. If the package supports deleting configs, use that.
"""

from datetime import datetime
from json import dumps

from boto3.session import Session
from botocore.exceptions import ClientError, NoCredentialsError
from dateutil.tz.tz import tzlocal


class EcsClient(object):
    def __init__(self, access_key_id=None, secret_access_key=None,
                 region=None, profile=None):
        session = Session(aws_access_key_id=access_key_id,
                          aws_secret_access_key=secret_access_key,
                          region_name=region,
                          profile_name=profile)
        self.boto = session.client(u'ecs')

    def describe_services(self, cluster_name, service_name):
        return self.boto.describe_services(
            cluster=cluster_name,
            services=[service_name]
        )

    def describe_task_definition(self, task_definition_arn):
        try:
            return self.boto.describe_task_definition(
                taskDefinition=task_definition_arn
            )
        except ClientError:
            raise UnknownTaskDefinitionError(
                u'Unknown task definition arn: %s' % task_definition_arn
            )

    def list_tasks(self, cluster_name, service_name):
        return self.boto.list_tasks(
            cluster=cluster_name,
            serviceName=service_name
        )

    def list_task_definitions(self, family_prefix, status='ACTIVE', sort='DESC'):
        response = self.boto.list_task_definitions(familyPrefix=family_prefix, status=status, sort=sort)
        return response['taskDefinitionArns']

    def describe_tasks(self, cluster_name, task_arns):
        return self.boto.describe_tasks(cluster=cluster_name, tasks=task_arns)

    def register_task_definition(self, family, containers, volumes, role_arn, cpu=False, memory=False, execution_role_arn=None,
                                 requires_compatibilities=[], network_mode='bridge'):
        fargate_td = {}
        if 'FARGATE' in requires_compatibilities:
            fargate_td = {
                'executionRoleArn': execution_role_arn or u'',
                'requiresCompatibilities': requires_compatibilities or [],
                'networkMode': network_mode or u'',
                'cpu': cpu,
                'memory': memory,
            }
        return self.boto.register_task_definition(
            family=family,
            containerDefinitions=containers,
            volumes=volumes,
            taskRoleArn=role_arn or u'',
            **fargate_td
        )

    def deregister_task_definition(self, task_definition_arn):
        return self.boto.deregister_task_definition(
            taskDefinition=task_definition_arn
        )

    def update_service(self, cluster, service, desired_count, task_definition):
        return self.boto.update_service(
            cluster=cluster,
            service=service,
            desiredCount=desired_count,
            taskDefinition=task_definition
        )

    def run_task(self, cluster, task_definition, count, started_by, overrides):
        return self.boto.run_task(
            cluster=cluster,
            taskDefinition=task_definition,
            count=count,
            startedBy=started_by,
            overrides=overrides
        )


class EcsService(dict):
    def __init__(self, cluster, service_definition=None, **kwargs):
        self._cluster = cluster
        super(EcsService, self).__init__(service_definition, **kwargs)

    def set_desired_count(self, desired_count):
        self[u'desiredCount'] = desired_count

    def set_task_definition(self, task_definition):
        self[u'taskDefinition'] = task_definition.arn

    @property
    def cluster(self):
        return self._cluster

    @property
    def name(self):
        return self.get(u'serviceName')

    @property
    def task_definition(self):
        return self.get(u'taskDefinition')

    @property
    def desired_count(self):
        return self.get(u'desiredCount')

    @property
    def deployment_created_at(self):
        for deployment in self.get(u'deployments'):
            if deployment.get(u'status') == u'PRIMARY':
                return deployment.get(u'createdAt')
        return datetime.now()

    @property
    def deployment_updated_at(self):
        for deployment in self.get(u'deployments'):
            if deployment.get(u'status') == u'PRIMARY':
                return deployment.get(u'updatedAt')
        return datetime.now()

    @property
    def errors(self):
        return self.get_warnings(
            since=self.deployment_updated_at
        )

    @property
    def older_errors(self):
        return self.get_warnings(
            since=self.deployment_created_at,
            until=self.deployment_updated_at
        )

    def get_warnings(self, since=None, until=None):
        since = since or self.deployment_created_at
        until = until or datetime.now(tz=tzlocal())
        errors = {}
        for event in self.get(u'events'):
            if u'unable' not in event[u'message']:
                continue
            if since < event[u'createdAt'] < until:
                errors[event[u'createdAt']] = event[u'message']
        return errors


class EcsTaskDefinition(dict):
    def __init__(self, task_definition=None, **kwargs):
        super(EcsTaskDefinition, self).__init__(task_definition, **kwargs)
        self._diff = []

    @property
    def containers(self):
        return self.get(u'containerDefinitions')

    @property
    def container_names(self):
        for container in self.get(u'containerDefinitions'):
            yield container[u'name']

    @property
    def volumes(self):
        return self.get(u'volumes')

    @property
    def arn(self):
        return self.get(u'taskDefinitionArn')

    @property
    def requires_compatibilities(self):
        return self.get(u'requiresCompatibilities')

    @property
    def execution_role_arn(self):
        return self.get(u'executionRoleArn')

    @property
    def network_mode(self):
        return self.get(u'networkMode')

    @property
    def cpu(self):
        return self.get(u'cpu')

    @property
    def memory(self):
        return self.get(u'memory')

    @property
    def family(self):
        return self.get(u'family')

    @property
    def role_arn(self):
        return self.get(u'taskRoleArn')

    @property
    def revision(self):
        return self.get(u'revision')

    @property
    def family_revision(self):
        return '%s:%d' % (self.get(u'family'), self.get(u'revision'))

    @property
    def diff(self):
        return self._diff

    def get_overrides(self):
        override = dict()
        overrides = []
        for diff in self.diff:
            if override.get('name') != diff.container:
                override = dict(name=diff.container)
                overrides.append(override)
            if diff.field == 'command':
                override['command'] = self.get_overrides_command(diff.value)
            elif diff.field == 'environment':
                override['environment'] = self.get_overrides_env(diff.value)
        return overrides

    @staticmethod
    def get_overrides_command(command):
        return command.split(' ')

    @staticmethod
    def get_overrides_env(env):
        return [{"name": e, "value": env[e]} for e in env]

    def set_images(self, tag=None, **images):
        self.validate_container_options(**images)
        for container in self.containers:
            if container[u'name'] in images:
                new_image = images[container[u'name']]
                diff = EcsTaskDefinitionDiff(
                    container=container[u'name'],
                    field=u'image',
                    value=new_image,
                    old_value=container[u'image']
                )
                self._diff.append(diff)
                container[u'image'] = new_image
            elif tag:
                image_definition = container[u'image'].rsplit(u':', 1)
                new_image = u'%s:%s' % (image_definition[0], tag.strip())
                diff = EcsTaskDefinitionDiff(
                    container=container[u'name'],
                    field=u'image',
                    value=new_image,
                    old_value=container[u'image']
                )
                self._diff.append(diff)
                container[u'image'] = new_image

    def set_commands(self, **commands):
        self.validate_container_options(**commands)
        for container in self.containers:
            if container[u'name'] in commands:
                new_command = commands[container[u'name']]
                diff = EcsTaskDefinitionDiff(
                    container=container[u'name'],
                    field=u'command',
                    value=new_command,
                    old_value=container.get(u'command')
                )
                self._diff.append(diff)
                container[u'command'] = [new_command]

    def set_environment(self, environment_list):
        environment = {}

        for env in environment_list:
            environment.setdefault(env[0], {})
            environment[env[0]][env[1]] = env[2]

        self.validate_container_options(**environment)
        for container in self.containers:
            if container[u'name'] in environment:
                self.apply_container_environment(
                    container=container,
                    new_environment=environment[container[u'name']]
                )

    def apply_container_environment(self, container, new_environment):
        old_environment = {
            env['name']: env['value'] for env in container.get(
                'environment',
                {}
            )
        }
        merged_environment = {var[0]: var[1] for var in new_environment}

        diff = EcsTaskDefinitionDiff(
            container[u'name'],
            u'environment',
            merged_environment,
            old_environment
        )
        self._diff.append(diff)

        container[u'environment'] = [
            {
                "name": e,
                "value": merged_environment[e]
            } for e in merged_environment
        ]

    def validate_container_options(self, **container_options):
        for container_name in container_options:
            if container_name not in self.container_names:
                raise UnknownContainerError(
                    u'Unknown container: %s' % container_name
                )

    def set_role_arn(self, role_arn):
        if role_arn:
            diff = EcsTaskDefinitionDiff(
                container=None,
                field=u'role_arn',
                value=role_arn,
                old_value=self[u'taskRoleArn']
            )
            self[u'taskRoleArn'] = role_arn
            self._diff.append(diff)


class EcsTaskDefinitionDiff(object):
    def __init__(self, container, field, value, old_value):
        self.container = container
        self.field = field
        self.value = value
        self.old_value = old_value

    def __repr__(self):
        if self.container:
            return u"Changed %s of container '%s' to: %s (was: %s)" % (
                self.field,
                self.container,
                dumps(self.value),
                dumps(self.old_value)
            )
        else:
            return u"Changed %s to: %s (was: %s)" % (
                self.field,
                dumps(self.value),
                dumps(self.old_value)
            )


class EcsAction(object):
    def __init__(self, client, cluster_name, service_name):
        self._client = client
        self._cluster_name = cluster_name
        self._service_name = service_name

        try:
            if service_name:
                self._service = self.get_service()
        except IndexError:
            raise EcsConnectionError(
                u'An error occurred when calling the DescribeServices '
                u'operation: Service not found.'
            )
        except ClientError as e:
            raise EcsConnectionError(str(e))
        except NoCredentialsError:
            raise EcsConnectionError(
                u'Unable to locate credentials. Configure credentials '
                u'by running "aws configure".'
            )

    def get_service(self):
        services_definition = self._client.describe_services(
            cluster_name=self._cluster_name,
            service_name=self._service_name
        )
        return EcsService(
            cluster=self._cluster_name,
            service_definition=services_definition[u'services'][0]
        )

    def get_current_task_definition(self, service):
        task_definition_payload = self._client.describe_task_definition(
            task_definition_arn=service.task_definition
        )
        task_definition = EcsTaskDefinition(
            task_definition=task_definition_payload[u'taskDefinition']
        )
        return task_definition

    def get_task_definition(self, task_definition):
        task_definition_payload = self._client.describe_task_definition(
            task_definition_arn=task_definition
        )
        task_definition = EcsTaskDefinition(
            task_definition=task_definition_payload[u'taskDefinition']
        )
        return task_definition

    def update_task_definition(self, task_definition):
        fargate_td = {}
        if task_definition.requires_compatibilities and 'FARGATE' in task_definition.requires_compatibilities:

            fargate_td = {
                'execution_role_arn': task_definition.execution_role_arn or u'',
                'requires_compatibilities': task_definition.requires_compatibilities or [],
                'network_mode': task_definition.network_mode or u'',
                'cpu' : task_definition.cpu or u'',
                'memory' : task_definition.memory or u'',

            }
        response = self._client.register_task_definition(
            family=task_definition.family,
            containers=task_definition.containers,
            volumes=task_definition.volumes,
            role_arn=task_definition.role_arn,

            **fargate_td
        )
        new_task_definition = EcsTaskDefinition(response[u'taskDefinition'])
        self._client.deregister_task_definition(task_definition.arn)
        return new_task_definition

    def update_service(self, service):
        response = self._client.update_service(
            cluster=service.cluster,
            service=service.name,
            desired_count=service.desired_count,
            task_definition=service.task_definition
        )
        return EcsService(self._cluster_name, response[u'service'])

    def is_deployed(self, service):
        if len(service[u'deployments']) != 1:
            return False
        running_tasks = self._client.list_tasks(
            cluster_name=service.cluster,
            service_name=service.name
        )
        if not running_tasks[u'taskArns']:
            return service.desired_count == 0
        running_count = self.get_running_tasks_count(
            service=service,
            task_arns=running_tasks[u'taskArns']
        )
        return service.desired_count == running_count

    def get_running_tasks_count(self, service, task_arns):
        running_count = 0
        tasks_details = self._client.describe_tasks(
            cluster_name=self._cluster_name,
            task_arns=task_arns
        )
        for task in tasks_details[u'tasks']:
            arn = task[u'taskDefinitionArn']
            status = task[u'lastStatus']
            if arn == service.task_definition and status == u'RUNNING':
                running_count += 1
        return running_count

    @property
    def client(self):
        return self._client

    @property
    def service(self):
        return self._service

    @property
    def cluster_name(self):
        return self._cluster_name

    @property
    def service_name(self):
        return self._service_name


class DeployAction(EcsAction):
    def deploy(self, task_definition):
        self._service.set_task_definition(task_definition)
        return self.update_service(self._service)


class ScaleAction(EcsAction):
    def scale(self, desired_count):
        self._service.set_desired_count(desired_count)
        return self.update_service(self._service)


class RunAction(EcsAction):
    def __init__(self, client, cluster_name):
        super(RunAction, self).__init__(client, cluster_name, None)
        self._client = client
        self._cluster_name = cluster_name
        self.started_tasks = []

    def run(self, task_definition, count, started_by):
        result = self._client.run_task(
            cluster=self._cluster_name,
            task_definition=task_definition.family_revision,
            count=count,
            started_by=started_by,
            overrides=dict(containerOverrides=task_definition.get_overrides())
        )
        self.started_tasks = result['tasks']
        return True


class EcsError(Exception):
    pass


class EcsConnectionError(EcsError):
    pass


class UnknownContainerError(EcsError):
    pass


class TaskPlacementError(EcsError):
    pass


class UnknownTaskDefinitionError(EcsError):
    pass



from setuptools import find_packages, setup

version = {}
with open("./cloudlift/version/__init__.py") as fp:
    exec(fp.read())

requirements = []
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='cloudlift',
    version=VERSION,
    packages=find_packages(),
    install_requires=requirements,
    description="Cloudlift makes it easier to launch dockerized services in AWS ECS",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/GetSimpl/cloudlift",
    zip_safe=True,
    entry_points='''
        [console_scripts]
        cloudlift=cloudlift:cli
    ''',
    python_requires='>=3.6',
)



import os

import botocore
from boto3 import client
from boto3.session import Session
from cloudlift.exceptions import UnrecoverableException

from cloudlift.config import get_account_id
from cloudlift.config.logging import log_bold, log_err


def do_mfa_login(mfa_code=None, region='ap-south-1'):
    username = get_username()
    if not mfa_code:
        mfa_code = input("MFA Code: ")
    mfa_arn = "arn:aws:iam::%s:mfa/%s" % (get_account_id(), username)

    log_bold("Using credentials for " + username)
    try:
        session_params = client('sts').get_session_token(
            DurationSeconds=900,
            SerialNumber=mfa_arn,
            TokenCode=str(mfa_code)
        )
        credentials = session_params['Credentials']
        os.environ['AWS_ACCESS_KEY_ID'] = credentials['AccessKeyId']
        os.environ['AWS_SECRET_ACCESS_KEY'] = credentials['SecretAccessKey']
        os.environ['AWS_SESSION_TOKEN'] = credentials['SessionToken']
        os.environ['AWS_DEFAULT_REGION'] = region
        return session_params
    except botocore.exceptions.ClientError as client_error:
        raise UnrecoverableException(str(client_error))


def get_mfa_session(mfa_code=None, region='ap-south-1'):
    username = get_username()
    if not mfa_code:
        mfa_code = input("MFA Code: ")
    mfa_arn = "arn:aws:iam::%s:mfa/%s" % (get_account_id(), username)

    log_bold("Using credentials for " + username)
    try:
        session_params = client('sts').get_session_token(
            DurationSeconds=900,
            SerialNumber=mfa_arn,
            TokenCode=str(mfa_code)
        )
        credentials = session_params['Credentials']
        return Session(
                aws_access_key_id=credentials['AccessKeyId'],
                aws_secret_access_key=credentials['SecretAccessKey'],
                aws_session_token=credentials['SessionToken'],
                region_name=region
        )
    except botocore.exceptions.ClientError as client_error:
        raise UnrecoverableException(str(client_error))


def get_username():
    return client('sts').get_caller_identity()['Arn'].split("user/")[1]



import os
import time

import boto3
import click
import requests
import urllib3
from mock import patch

from cloudlift.config import ServiceConfiguration, VERSION
from cloudlift.deployment.service_creator import ServiceCreator
from cloudlift.deployment.service_updater import ServiceUpdater


def setup_module(module):
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def mocked_service_config(cls, *args, **kwargs):
    return None

def mocked_fargate_service_config(cls, *args, **kwargs):
    return {
        "cloudlift_version": VERSION,
        "services": {
            "DummyFargateService": {
                "command": None,
                "fargate": {
                    "cpu": 256,
                    "memory": 512
                },
                "http_interface": {
                    "container_port": 80,
                    "internal": False,
                    "restrict_access_to": [
                        "0.0.0.0/0"
                    ],
                    "health_check_path": "/elb-check"
                },
                "memory_reservation": 512
            }
        }
    }


environment_name = 'staging'
service_name = 'dummy'
fargate_service_name = 'dummy-fargate'

def test_cloudlift_can_deploy_to_ec2(keep_resources):
    cfn_client = boto3.client('cloudformation')
    stack_name = f'{service_name}-{environment_name}'
    cfn_client.delete_stack(StackName=stack_name)
    print("initiated delete")
    waiter = cfn_client.get_waiter('stack_delete_complete')
    waiter.wait(StackName=stack_name)
    print("completed delete")
    config_path = '/'.join([environment_name, service_name, 'env.properties'])
    os.chdir('./test/dummy')
    print("adding configuration to parameter store")
    ssm_client = boto3.client('ssm')
    ssm_client.put_parameter(
        Name=f"/{environment_name}/{service_name}/PORT",
        Value="80",
        Type="SecureString",
        KeyId='alias/aws/ssm', Overwrite=True
    )
    ssm_client.put_parameter(
        Name=f"/{environment_name}/{service_name}/LABEL",
        Value="Demo",
        Type="SecureString",
        KeyId='alias/aws/ssm',
        Overwrite=True
    )
    with patch.object(ServiceConfiguration, 'edit_config',
                      new=mocked_service_config):
        ServiceCreator(service_name, environment_name,).create()
    ServiceUpdater(service_name, environment_name, None).run()
    outputs = cfn_client.describe_stacks(
        StackName=stack_name
    )['Stacks'][0]['Outputs']
    service_url = [
        x for x in outputs if x["OutputKey"] == "DummyURL"
    ][0]['OutputValue']
    content_matched = wait_until(
        lambda: match_page_content(
            service_url,
            'This is dummy app. Label: Demo'
        ), 60)
    os.chdir('../../')
    assert content_matched
    if not keep_resources:
        cfn_client.delete_stack(StackName=stack_name)


def test_cloudlift_can_deploy_to_fargate(keep_resources):
    cfn_client = boto3.client('cloudformation')
    stack_name = f'{fargate_service_name}-{environment_name}'
    cfn_client.delete_stack(StackName=stack_name)
    print("initiated delete of " + stack_name)
    waiter = cfn_client.get_waiter('stack_delete_complete')
    waiter.wait(StackName=stack_name)
    print("completed delete")
    config_path = '/'.join([environment_name, fargate_service_name, 'env.properties'])
    os.chdir('./test/dummy')
    print("adding configuration to parameter store")
    ssm_client = boto3.client('ssm')
    ssm_client.put_parameter(
        Name=f"/{environment_name}/{fargate_service_name}/PORT",
        Value="80",
        Type="SecureString",
        KeyId='alias/aws/ssm', Overwrite=True
    )
    ssm_client.put_parameter(
        Name=f"/{environment_name}/{fargate_service_name}/LABEL",
        Value="Demo",
        Type="SecureString",
        KeyId='alias/aws/ssm',
        Overwrite=True
    )
    with patch.object(ServiceConfiguration, 'edit_config',
                     new=mocked_fargate_service_config):
        with patch.object(ServiceConfiguration, 'get_config',
                          new=mocked_fargate_service_config):
            ServiceCreator(fargate_service_name, environment_name,).create()
    ServiceUpdater(fargate_service_name, environment_name, None).run()
    outputs = cfn_client.describe_stacks(
        StackName=stack_name
    )['Stacks'][0]['Outputs']
    service_url = [
        x for x in outputs if x["OutputKey"] == "DummyFargateServiceURL"
    ][0]['OutputValue']
    content_matched = wait_until(
        lambda: match_page_content(
            service_url,
            'This is dummy app. Label: Demo'
        ), 60)
    os.chdir('../../')
    assert content_matched
    if not keep_resources:
        cfn_client.delete_stack(StackName=stack_name)


def match_page_content(service_url, content_expected):
    page_content = requests.get(service_url, verify=False).text
    print("page_content: " + str(page_content))
    return page_content == content_expected


def wait_until(predicate, timeout, period=1, *args, **kwargs):
    mustend = time.time() + timeout
    while time.time() < mustend:
        if predicate(*args, **kwargs):
            return True
        print("sleeping and gonna retry...")
        time.sleep(period)
    return False




from subprocess import call

from cloudlift.config import get_client_for
from cloudlift.config import get_cluster_name, get_service_stack_name
from cloudlift.config.logging import log, log_bold, log_err, log_intent, log_warning


class ServiceInformationFetcher(object):
    def __init__(self, name, environment):
        self.name = name
        self.environment = environment
        self.cluster_name = get_cluster_name(environment)
        self.ecs_client = get_client_for('ecs', self.environment)
        self.ec2_client = get_client_for('ec2', self.environment)
        self.init_stack_info()

    def init_stack_info(self):
        self.stack_name = get_service_stack_name(self.environment, self.name)
        try:
            stack = get_client_for(
                'cloudformation',
                self.environment).describe_stacks(
                    StackName=self.stack_name
                )['Stacks'][0]
            service_name_list = list(
                filter(
                    lambda x: x['OutputKey'].endswith('EcsServiceName'),
                    stack['Outputs']
                )
            )
            self.ecs_service_names = [
                svc_name['OutputValue'] for svc_name in service_name_list
            ]
        except Exception:
            self.ecs_service_names = []
            log_warning("Could not determine services.")

    def get_current_version(self):
        commit_sha = self._fetch_current_task_definition_tag()
        if commit_sha is None or commit_sha == 'dirty':
            log_warning("Currently deployed tag could not be found or is dirty,\
resetting to master")
            commit_sha = "master"
        return commit_sha

    def log_ips(self):
        for service in self.ecs_service_names:
            task_arns = self.ecs_client.list_tasks(
                cluster=self.cluster_name,
                serviceName=service
            )['taskArns']
            tasks = self.ecs_client.describe_tasks(
                cluster=self.cluster_name,
                tasks=task_arns
            )['tasks']
            container_instance_arns = [
                task['containerInstanceArn'] for task in tasks
            ]
            container_instances = self.ecs_client.describe_container_instances(
                cluster=self.cluster_name,
                containerInstances=container_instance_arns
            )['containerInstances']
            ecs_instance_ids = [
                container['ec2InstanceId'] for container in container_instances
            ]
            ec2_reservations = self.ec2_client.describe_instances(
                InstanceIds=ecs_instance_ids
            )['Reservations']
            log_bold(service,)
            for reservation in ec2_reservations:
                instances = reservation['Instances']
                ips = [instance['PrivateIpAddress'] for instance in instances]
                [log_intent(ip) for ip in ips]
            log("")

    def get_instance_ids(self):
        instance_ids = {}
        for service in self.ecs_service_names:
            task_arns = self.ecs_client.list_tasks(
                cluster=self.cluster_name,
                serviceName=service
            )['taskArns']
            tasks = self.ecs_client.describe_tasks(
                cluster=self.cluster_name,
                tasks=task_arns
            )['tasks']
            container_instance_arns = [
                task['containerInstanceArn'] for task in tasks
            ]
            container_instances = self.ecs_client.describe_container_instances(
                cluster=self.cluster_name,
                containerInstances=container_instance_arns
            )['containerInstances']
            service_instance_ids = [
                container['ec2InstanceId'] for container in container_instances
            ]
            instance_ids[service] = service_instance_ids
        return instance_ids

    def get_version(self, short):
        commit_sha = self._fetch_current_task_definition_tag()
        if commit_sha is None:
            log_err("Current task definition tag could not be found. \
Is it deployed?")
        elif commit_sha == "dirty":
            log("Dirty version is deployed. Commit information could not be \
fetched.")
        else:
            log("Currently deployed version: " + commit_sha)
            if not short:
                log("Running `git fetch --all`")
                call(["git", "fetch", "--all"])
                log_bold("Commit Info:")
                call([
                    "git",
                    "--no-pager",
                    "show",
                    "-s",
                    "--format=medium",
                    commit_sha
                ])
                log_bold("Branch Info:")
                call(["git", "branch", "-r", "--contains", commit_sha])
                log("")

    def _fetch_current_task_definition_tag(self):
        try:
            service = self.ecs_service_names[0]
            task_arns = self.ecs_client.list_tasks(
                cluster=self.cluster_name,
                serviceName=service
            )['taskArns']
            tasks = self.ecs_client.describe_tasks(
                cluster=self.cluster_name,
                tasks=task_arns
            )['tasks']
            task_definition_arns = tasks[0]['taskDefinitionArn']
            task_definition = self.ecs_client.describe_task_definition(
                taskDefinition=task_definition_arns
            )
            image = task_definition['taskDefinition']['containerDefinitions'][0]['image']
            commit_sha = image.split('-repo:')[1]
            return commit_sha
        except Exception:
            return None



import json
import re
import uuid

import boto3
from botocore.exceptions import ClientError
from cloudlift.exceptions import UnrecoverableException
from cloudlift.config import get_client_for

from awacs.aws import PolicyDocument, Statement, Allow, Principal
from awacs.sts import AssumeRole
from cfn_flip import to_yaml
from stringcase import pascalcase
from troposphere import GetAtt, Output, Parameter, Ref, Sub
from troposphere.cloudwatch import Alarm, MetricDimension
from troposphere.ec2 import SecurityGroup
from troposphere.ecs import (AwsvpcConfiguration, ContainerDefinition,
                             DeploymentConfiguration, Environment,
                             LoadBalancer, LogConfiguration,
                             NetworkConfiguration, PlacementStrategy,
                             PortMapping, Service, TaskDefinition)
from troposphere.elasticloadbalancingv2 import Action, Certificate, Listener
from troposphere.elasticloadbalancingv2 import LoadBalancer as ALBLoadBalancer
from troposphere.elasticloadbalancingv2 import (Matcher, RedirectConfig,
                                                TargetGroup,
                                                TargetGroupAttribute)
from troposphere.iam import Role

from cloudlift.config import region as region_service
from cloudlift.config import get_account_id
from cloudlift.config import DecimalEncoder
from cloudlift.config import get_service_stack_name
from cloudlift.deployment.deployer import build_config
from cloudlift.deployment.ecs import DeployAction, EcsClient
from cloudlift.config.logging import log, log_bold
from cloudlift.deployment.service_information_fetcher import ServiceInformationFetcher
from cloudlift.deployment.template_generator import TemplateGenerator


class ServiceTemplateGenerator(TemplateGenerator):
    PLACEMENT_STRATEGIES = [
        PlacementStrategy(
            Type='spread',
            Field='attribute:ecs.availability-zone'
        ),
        PlacementStrategy(
            Type='spread',
            Field='instanceId'
        )]
    LAUNCH_TYPE_FARGATE = 'FARGATE'
    LAUNCH_TYPE_EC2 = 'EC2'

    def __init__(self, service_configuration, environment_stack):
        super(ServiceTemplateGenerator, self).__init__(
            service_configuration.environment
        )
        self._derive_configuration(service_configuration)
        self.env_sample_file_path = './env.sample'
        self.environment_stack = environment_stack
        self.current_version = ServiceInformationFetcher(
            self.application_name, self.env).get_current_version()
        self.bucket_name = 'cloudlift-service-template'
        self.environment = service_configuration.environment
        self.client = get_client_for('s3', self.environment)

    def _derive_configuration(self, service_configuration):
        self.application_name = service_configuration.service_name
        self.configuration = service_configuration.get_config()

    def generate_service(self):
        self._add_service_parameters()
        self._add_service_outputs()
        self._fetch_current_desired_count()
        self._add_ecs_service_iam_role()
        self._add_cluster_services()

        key = uuid.uuid4().hex + '.yml'
        if len(to_yaml(self.template.to_json())) > 51000:
            try:
                self.client.put_object(
                    Body=to_yaml(self.template.to_json()),
                    Bucket=self.bucket_name,
                    Key=key,
                )
                template_url = f'https://{self.bucket_name}.s3.amazonaws.com/{key}'
                return template_url, 'TemplateURL', key
            except ClientError as boto_client_error:
                error_code = boto_client_error.response['Error']['Code']
                if error_code == 'AccessDenied':
                    raise UnrecoverableException(f'Unable to store cloudlift service template in S3 bucket at {self.bucket_name}')
                else:
                    raise boto_client_error
        else:
            return to_yaml(self.template.to_json()), 'TemplateBody', ''

    def _add_cluster_services(self):
        for ecs_service_name, config in self.configuration['services'].items():
            self._add_service(ecs_service_name, config)

    def _add_service_alarms(self, svc):
        ecs_high_cpu_alarm = Alarm(
            'EcsHighCPUAlarm' + str(svc.name),
            EvaluationPeriods=1,
            Dimensions=[
                MetricDimension(
                    Name='ClusterName',
                    Value=self.cluster_name
                ),
                MetricDimension(
                    Name='ServiceName',
                    Value=GetAtt(svc, 'Name')
                )],
            AlarmActions=[Ref(self.notification_sns_arn)],
            OKActions=[Ref(self.notification_sns_arn)],
            AlarmDescription='Alarm if CPU too high or metric disappears \
indicating instance is down',
            Namespace='AWS/ECS',
            Period=300,
            ComparisonOperator='GreaterThanThreshold',
            Statistic='Average',
            Threshold='80',
            MetricName='CPUUtilization'
        )
        self.template.add_resource(ecs_high_cpu_alarm)
        ecs_high_memory_alarm = Alarm(
            'EcsHighMemoryAlarm' + str(svc.name),
            EvaluationPeriods=1,
            Dimensions=[
                MetricDimension(
                    Name='ClusterName',
                    Value=self.cluster_name
                ),
                MetricDimension(
                    Name='ServiceName',
                    Value=GetAtt(svc, 'Name')
                )
            ],
            AlarmActions=[Ref(self.notification_sns_arn)],
            OKActions=[Ref(self.notification_sns_arn)],
            AlarmDescription='Alarm if memory too high or metric \
disappears indicating instance is down',
            Namespace='AWS/ECS',
            Period=300,
            ComparisonOperator='GreaterThanThreshold',
            Statistic='Average',
            Threshold='80',
            MetricName='MemoryUtilization'
        )
        self.template.add_resource(ecs_high_memory_alarm)
        # How to add service task count alarm
        # http://docs.aws.amazon.com/AmazonECS/latest/developerguide/cloudwatch-metrics.html#cw_running_task_count
        ecs_no_running_tasks_alarm = Alarm(
            'EcsNoRunningTasksAlarm' + str(svc.name),
            EvaluationPeriods=1,
            Dimensions=[
                MetricDimension(
                    Name='ClusterName',
                    Value=self.cluster_name
                ),
                MetricDimension(
                    Name='ServiceName',
                    Value=GetAtt(svc, 'Name')
                )
            ],
            AlarmActions=[Ref(self.notification_sns_arn)],
            OKActions=[Ref(self.notification_sns_arn)],
            AlarmDescription='Alarm if the task count goes to zero, denoting \
service is down',
            Namespace='AWS/ECS',
            Period=60,
            ComparisonOperator='LessThanThreshold',
            Statistic='SampleCount',
            Threshold='1',
            MetricName='CPUUtilization',
            TreatMissingData='breaching'
        )
        self.template.add_resource(ecs_no_running_tasks_alarm)

    def _add_service(self, service_name, config):
        launch_type = self.LAUNCH_TYPE_FARGATE if 'fargate' in config else self.LAUNCH_TYPE_EC2
        env_config = build_config(
            self.env,
            self.application_name,
            self.env_sample_file_path
        )
        container_definition_arguments = {
            "Environment": [
                Environment(Name=k, Value=v) for (k, v) in env_config
            ],
            "Name": service_name + "Container",
            "Image": self.ecr_image_uri + ':' + self.current_version,
            "Essential": 'true',
            "LogConfiguration": self._gen_log_config(service_name),
            "MemoryReservation": int(config['memory_reservation']),
            "Cpu": 0
        }

        if 'http_interface' in config:
            container_definition_arguments['PortMappings'] = [
                PortMapping(
                    ContainerPort=int(
                        config['http_interface']['container_port']
                    )
                )
            ]

        if config['command'] is not None:
            container_definition_arguments['Command'] = [config['command']]

        cd = ContainerDefinition(**container_definition_arguments)

        task_role = self.template.add_resource(Role(
            service_name + "Role",
            AssumeRolePolicyDocument=PolicyDocument(
                Statement=[
                    Statement(
                        Effect=Allow,
                        Action=[AssumeRole],
                        Principal=Principal("Service", ["ecs-tasks.amazonaws.com"])
                    )
                ]
            )
        ))

        launch_type_td = {}
        if launch_type == self.LAUNCH_TYPE_FARGATE:
            launch_type_td = {
                'RequiresCompatibilities': ['FARGATE'],
                'ExecutionRoleArn': boto3.resource('iam').Role('ecsTaskExecutionRole').arn,
                'NetworkMode': 'awsvpc',
                'Cpu': str(config['fargate']['cpu']),
                'Memory': str(config['fargate']['memory'])
            }

        td = TaskDefinition(
            service_name + "TaskDefinition",
            Family=service_name + "Family",
            ContainerDefinitions=[cd],
            TaskRoleArn=Ref(task_role),
            **launch_type_td
        )

        self.template.add_resource(td)
        desired_count = self._get_desired_task_count_for_service(service_name)
        deployment_configuration = DeploymentConfiguration(
            MinimumHealthyPercent=100,
            MaximumPercent=200
        )
        if 'http_interface' in config:
            alb, lb, service_listener, alb_sg = self._add_alb(cd, service_name, config, launch_type)

            if launch_type == self.LAUNCH_TYPE_FARGATE:
                # if launch type is ec2, then services inherit the ec2 instance security group
                # otherwise, we need to specify a security group for the service
                service_security_group = SecurityGroup(
                    pascalcase("FargateService" + self.env + service_name),
                    GroupName=pascalcase("FargateService" + self.env + service_name),
                    SecurityGroupIngress=[{
                        'IpProtocol': 'TCP',
                        'SourceSecurityGroupId': Ref(alb_sg),
                        'ToPort': int(config['http_interface']['container_port']),
                        'FromPort': int(config['http_interface']['container_port']),
                    }],
                    VpcId=Ref(self.vpc),
                    GroupDescription=pascalcase("FargateService" + self.env + service_name)
                )
                self.template.add_resource(service_security_group)

                launch_type_svc = {
                    'NetworkConfiguration': NetworkConfiguration(
                        AwsvpcConfiguration=AwsvpcConfiguration(
                            Subnets=[
                                Ref(self.private_subnet1),
                                Ref(self.private_subnet2)
                            ],
                            SecurityGroups=[
                                Ref(service_security_group)
                            ]
                        )
                    )
                }
            else:
                launch_type_svc = {
                    'Role': Ref(self.ecs_service_role),
                    'PlacementStrategies': self.PLACEMENT_STRATEGIES
                }
            svc = Service(
                service_name,
                LoadBalancers=[lb],
                Cluster=self.cluster_name,
                TaskDefinition=Ref(td),
                DesiredCount=desired_count,
                DependsOn=service_listener.title,
                LaunchType=launch_type,
                **launch_type_svc,
            )
            self.template.add_output(
                Output(
                    service_name + 'EcsServiceName',
                    Description='The ECS name which needs to be entered',
                    Value=GetAtt(svc, 'Name')
                )
            )
            self.template.add_output(
                Output(
                    service_name + "URL",
                    Description="The URL at which the service is accessible",
                    Value=Sub("https://${" + alb.name + ".DNSName}")
                )
            )
            self.template.add_resource(svc)
        else:
            launch_type_svc = {}
            if launch_type == self.LAUNCH_TYPE_FARGATE:
                # if launch type is ec2, then services inherit the ec2 instance security group
                # otherwise, we need to specify a security group for the service
                service_security_group = SecurityGroup(
                    pascalcase("FargateService" + self.env + service_name),
                    GroupName=pascalcase("FargateService" + self.env + service_name),
                    SecurityGroupIngress=[],
                    VpcId=Ref(self.vpc),
                    GroupDescription=pascalcase("FargateService" + self.env + service_name)
                )
                self.template.add_resource(service_security_group)
                launch_type_svc = {
                    'NetworkConfiguration': NetworkConfiguration(
                        AwsvpcConfiguration=AwsvpcConfiguration(
                            Subnets=[
                                Ref(self.private_subnet1),
                                Ref(self.private_subnet2)
                            ],
                            SecurityGroups=[
                                Ref(service_security_group)
                            ]
                        )
                    )
                }
            else:
                launch_type_svc = {
                    'PlacementStrategies': self.PLACEMENT_STRATEGIES
                }
            svc = Service(
                service_name,
                Cluster=self.cluster_name,
                TaskDefinition=Ref(td),
                DesiredCount=desired_count,
                DeploymentConfiguration=deployment_configuration,
                LaunchType=launch_type,
                **launch_type_svc
            )
            self.template.add_output(
                Output(
                    service_name + 'EcsServiceName',
                    Description='The ECS name which needs to be entered',
                    Value=GetAtt(svc, 'Name')
                )
            )
            self.template.add_resource(svc)
        self._add_service_alarms(svc)

    def _gen_log_config(self, service_name):
        return LogConfiguration(
            LogDriver="awslogs",
            Options={
                'awslogs-stream-prefix': service_name,
                'awslogs-group': '-'.join([self.env, 'logs']),
                'awslogs-region': self.region
            }
        )

    def _add_alb(self, cd, service_name, config, launch_type):
        sg_name = 'SG' + self.env + service_name
        svc_alb_sg = SecurityGroup(
            re.sub(r'\W+', '', sg_name),
            GroupName=self.env + '-' + service_name,
            SecurityGroupIngress=self._generate_alb_security_group_ingress(
                config
            ),
            VpcId=Ref(self.vpc),
            GroupDescription=Sub(service_name + "-alb-sg")
        )
        self.template.add_resource(svc_alb_sg)
        alb_name = service_name + pascalcase(self.env)
        if config['http_interface']['internal']:
            alb_subnets = [
                Ref(self.private_subnet1),
                Ref(self.private_subnet2)
            ]
            scheme = "internal"
            alb_name += 'Internal'
            alb_name = alb_name[:32]
            alb = ALBLoadBalancer(
                'ALB' + service_name,
                Subnets=alb_subnets,
                SecurityGroups=[
                    self.alb_security_group,
                    Ref(svc_alb_sg)
                ],
                Name=alb_name,
                Tags=[
                    {'Value': alb_name, 'Key': 'Name'}
                ],
                Scheme=scheme
            )
        else:
            alb_subnets = [
                Ref(self.public_subnet1),
                Ref(self.public_subnet2)
            ]
            alb_name = alb_name[:32]
            alb = ALBLoadBalancer(
                'ALB' + service_name,
                Subnets=alb_subnets,
                SecurityGroups=[
                    self.alb_security_group,
                    Ref(svc_alb_sg)
                ],
                Name=alb_name,
                Tags=[
                    {'Value': alb_name, 'Key': 'Name'}
                ]
            )

        self.template.add_resource(alb)

        target_group_name = "TargetGroup" + service_name
        health_check_path = config['http_interface']['health_check_path'] if 'health_check_path' in config['http_interface'] else "/elb-check"
        if config['http_interface']['internal']:
            target_group_name = target_group_name + 'Internal'

        target_group_config = {}
        if launch_type == self.LAUNCH_TYPE_FARGATE:
            target_group_config['TargetType'] = 'ip'

        service_target_group = TargetGroup(
            target_group_name,
            HealthCheckPath=health_check_path,
            HealthyThresholdCount=2,
            HealthCheckIntervalSeconds=30,
            TargetGroupAttributes=[
                TargetGroupAttribute(
                    Key='deregistration_delay.timeout_seconds',
                    Value='30'
                )
            ],
            VpcId=Ref(self.vpc),
            Protocol="HTTP",
            Matcher=Matcher(HttpCode="200-399"),
            Port=int(config['http_interface']['container_port']),
            HealthCheckTimeoutSeconds=10,
            UnhealthyThresholdCount=3,
            **target_group_config
        )

        self.template.add_resource(service_target_group)
        # Note: This is a ECS Loadbalancer definition. Not an ALB.
        # Defining this causes the target group to add a target to the correct
        # port in correct ECS cluster instance for the service container.
        lb = LoadBalancer(
            ContainerName=cd.Name,
            TargetGroupArn=Ref(service_target_group),
            ContainerPort=int(config['http_interface']['container_port'])
        )
        target_group_action = Action(
            TargetGroupArn=Ref(target_group_name),
            Type="forward"
        )
        service_listener = self._add_service_listener(
            service_name,
            target_group_action,
            alb,
            config['http_interface']['internal']
        )
        self._add_alb_alarms(service_name, alb)
        return alb, lb, service_listener, svc_alb_sg

    def _add_service_listener(self, service_name, target_group_action,
                              alb, internal):
        ssl_cert = Certificate(CertificateArn=self.ssl_certificate_arn)
        service_listener = Listener(
            "SslLoadBalancerListener" + service_name,
            Protocol="HTTPS",
            DefaultActions=[target_group_action],
            LoadBalancerArn=Ref(alb),
            Port=443,
            Certificates=[ssl_cert],
            SslPolicy="ELBSecurityPolicy-FS-1-2-Res-2019-08"
        )
        self.template.add_resource(service_listener)
        if internal:
            # Allow HTTP traffic on internal services
            http_service_listener = Listener(
                "LoadBalancerListener" + service_name,
                Protocol="HTTP",
                DefaultActions=[target_group_action],
                LoadBalancerArn=Ref(alb),
                Port=80
            )
            self.template.add_resource(http_service_listener)
        else:
            # Redirect HTTP to HTTPS on external services
            redirection_config = RedirectConfig(
                StatusCode='HTTP_301',
                Protocol='HTTPS',
                Port='443'
            )
            http_redirection_action = Action(
                RedirectConfig=redirection_config,
                Type="redirect"
            )
            http_redirection_listener = Listener(
                "LoadBalancerRedirectionListener" + service_name,
                Protocol="HTTP",
                DefaultActions=[http_redirection_action],
                LoadBalancerArn=Ref(alb),
                Port=80
            )
            self.template.add_resource(http_redirection_listener)
        return service_listener

    def _add_alb_alarms(self, service_name, alb):
        unhealthy_alarm = Alarm(
            'ElbUnhealthyHostAlarm' + service_name,
            EvaluationPeriods=1,
            Dimensions=[
                MetricDimension(
                    Name='LoadBalancer',
                    Value=GetAtt(alb, 'LoadBalancerFullName')
                )
            ],
            AlarmActions=[Ref(self.notification_sns_arn)],
            OKActions=[Ref(self.notification_sns_arn)],
            AlarmDescription='Triggers if any host is marked unhealthy',
            Namespace='AWS/ApplicationELB',
            Period=60,
            ComparisonOperator='GreaterThanOrEqualToThreshold',
            Statistic='Sum',
            Threshold='1',
            MetricName='UnHealthyHostCount',
            TreatMissingData='notBreaching'
        )
        self.template.add_resource(unhealthy_alarm)
        rejected_connections_alarm = Alarm(
            'ElbRejectedConnectionsAlarm' + service_name,
            EvaluationPeriods=1,
            Dimensions=[
                MetricDimension(
                    Name='LoadBalancer',
                    Value=GetAtt(alb, 'LoadBalancerFullName')
                )
            ],
            AlarmActions=[Ref(self.notification_sns_arn)],
            OKActions=[Ref(self.notification_sns_arn)],
            AlarmDescription='Triggers if load balancer has \
rejected connections because the load balancer \
had reached its maximum number of connections.',
            Namespace='AWS/ApplicationELB',
            Period=60,
            ComparisonOperator='GreaterThanOrEqualToThreshold',
            Statistic='Sum',
            Threshold='1',
            MetricName='RejectedConnectionCount',
            TreatMissingData='notBreaching'
        )
        self.template.add_resource(rejected_connections_alarm)
        http_code_elb5xx_alarm = Alarm(
            'ElbHTTPCodeELB5xxAlarm' + service_name,
            EvaluationPeriods=1,
            Dimensions=[
                MetricDimension(
                    Name='LoadBalancer',
                    Value=GetAtt(alb, 'LoadBalancerFullName')
                )
            ],
            AlarmActions=[Ref(self.notification_sns_arn)],
            OKActions=[Ref(self.notification_sns_arn)],
            AlarmDescription='Triggers if 5xx response originated \
from load balancer',
            Namespace='AWS/ApplicationELB',
            Period=60,
            ComparisonOperator='GreaterThanOrEqualToThreshold',
            Statistic='Sum',
            Threshold='3',
            MetricName='HTTPCode_ELB_5XX_Count',
            TreatMissingData='notBreaching'
        )
        self.template.add_resource(http_code_elb5xx_alarm)

    def _generate_alb_security_group_ingress(self, config):
        ingress_rules = []
        for access_ip in config['http_interface']['restrict_access_to']:
            if access_ip.find('/') == -1:
                access_ip = access_ip + '/32'
            ingress_rules.append({
                'ToPort': 80,
                'IpProtocol': 'TCP',
                'FromPort': 80,
                'CidrIp': access_ip
            })
            ingress_rules.append({
                'ToPort': 443,
                'IpProtocol': 'TCP',
                'FromPort': 443,
                'CidrIp': access_ip
            })
        return ingress_rules

    def _add_ecs_service_iam_role(self):
        role_name = Sub('ecs-svc-${AWS::StackName}-${AWS::Region}')
        assume_role_policy = {
            u'Statement': [
                {
                    u'Action': [u'sts:AssumeRole'],
                    u'Effect': u'Allow',
                    u'Principal': {
                        u'Service': [u'ecs.amazonaws.com']
                    }
                }
            ]
        }
        self.ecs_service_role = Role(
            'ECSServiceRole',
            Path='/',
            ManagedPolicyArns=[
                'arn:aws:iam::aws:policy/service-role/AmazonEC2ContainerServiceRole'
            ],
            RoleName=role_name,
            AssumeRolePolicyDocument=assume_role_policy
        )
        self.template.add_resource(self.ecs_service_role)

    def _add_service_outputs(self):
        self.template.add_output(Output(
            "CloudliftOptions",
            Description="Options used with cloudlift when \
building this service",
            Value=json.dumps(
                self.configuration,
                cls=DecimalEncoder
            )
        ))
        self._add_stack_outputs()

    def _add_service_parameters(self):
        self.notification_sns_arn = Parameter(
            "NotificationSnsArn",
            Description='',
            Type="String",
            Default=self.notifications_arn)
        self.template.add_parameter(self.notification_sns_arn)
        self.vpc = Parameter(
            "VPC",
            Description='',
            Type="AWS::EC2::VPC::Id",
            Default=list(
                filter(
                    lambda x: x['OutputKey'] == "VPC",
                    self.environment_stack['Outputs']
                )
            )[0]['OutputValue']
        )
        self.template.add_parameter(self.vpc)
        self.public_subnet1 = Parameter(
            "PublicSubnet1",
            Description='',
            Type="AWS::EC2::Subnet::Id",
            Default=list(
                filter(
                    lambda x: x['OutputKey'] == "PublicSubnet1",
                    self.environment_stack['Outputs']
                )
            )[0]['OutputValue']
        )
        self.template.add_parameter(self.public_subnet1)
        self.public_subnet2 = Parameter(
            "PublicSubnet2",
            Description='',
            Type="AWS::EC2::Subnet::Id",
            Default=list(
                filter(
                    lambda x: x['OutputKey'] == "PublicSubnet2",
                    self.environment_stack['Outputs']
                )
            )[0]['OutputValue']
        )
        self.template.add_parameter(self.public_subnet2)
        self.private_subnet1 = Parameter(
            "PrivateSubnet1",
            Description='',
            Type="AWS::EC2::Subnet::Id",
            Default=list(
                filter(
                    lambda x: x['OutputKey'] == "PrivateSubnet1",
                    self.environment_stack['Outputs']
                )
            )[0]['OutputValue']
        )
        self.template.add_parameter(self.private_subnet1)
        self.private_subnet2 = Parameter(
            "PrivateSubnet2",
            Description='',
            Type="AWS::EC2::Subnet::Id",
            Default=list(
                filter(
                    lambda x: x['OutputKey'] == "PrivateSubnet2",
                    self.environment_stack['Outputs']
                )
            )[0]['OutputValue']
        )
        self.template.add_parameter(self.private_subnet2)
        self.template.add_parameter(Parameter(
            "Environment",
            Description='',
            Type="String",
            Default="production"
        ))
        self.alb_security_group = list(
            filter(
                lambda x: x['OutputKey'] == "SecurityGroupAlb",
                self.environment_stack['Outputs']
            )
        )[0]['OutputValue']

    def _fetch_current_desired_count(self):
        stack_name = get_service_stack_name(self.env, self.application_name)
        self.desired_counts = {}
        try:
            stack = region_service.get_client_for(
                'cloudformation',
                self.env
            ).describe_stacks(StackName=stack_name)['Stacks'][0]
            ecs_service_outputs = filter(
                lambda x: x['OutputKey'].endswith('EcsServiceName'),
                stack['Outputs']
            )
            ecs_service_names = []
            for service_name in ecs_service_outputs:
                ecs_service_names.append({
                    "key": service_name['OutputKey'],
                    "value": service_name['OutputValue']
                })
            ecs_client = EcsClient(None, None, self.region)
            for service_name in ecs_service_names:
                deployment = DeployAction(
                    ecs_client,
                    self.cluster_name,
                    service_name["value"]
                )
                actual_service_name = service_name["key"]. \
                    replace("EcsServiceName", "")
                self.desired_counts[actual_service_name] = deployment. \
                    service.desired_count
            log("Existing service counts: " + str(self.desired_counts))
        except Exception:
            log_bold("Could not find existing services.")

    def _get_desired_task_count_for_service(self, service_name):
        if service_name in self.desired_counts:
            return self.desired_counts[service_name]
        else:
            return 0

    @property
    def ecr_image_uri(self):
        return str(self.account_id) + ".dkr.ecr." + \
               self.region + ".amazonaws.com/" + \
               self.repo_name

    @property
    def account_id(self):
        return get_account_id()

    @property
    def repo_name(self):
        return self.application_name + '-repo'

    @property
    def notifications_arn(self):
        """
        Get the SNS arn either from service configuration or the cluster
        """
        if 'notifications_arn' in self.configuration:
            return self.configuration['notifications_arn']
        else:
            return TemplateGenerator.notifications_arn.fget(self)



import base64
import multiprocessing
import os
import subprocess
import boto3
from time import sleep

from botocore.exceptions import ClientError

from cloudlift.deployment.ecr_client import EcrClient
from cloudlift.exceptions import UnrecoverableException
from stringcase import spinalcase, capitalcase

from cloudlift.config import get_account_id
from cloudlift.config import (get_client_for,
                              get_region_for_environment)
from cloudlift.config import get_cluster_name, get_service_stack_name
from cloudlift.deployment import deployer
from cloudlift.config.logging import log_bold, log_err, log_intent, log_warning
from cloudlift.deployment.ecs import DeployAction

DEPLOYMENT_COLORS = ['blue', 'magenta', 'white', 'cyan']


class ServiceUpdater(object):
    def __init__(self, name, environment, env_sample_file, version=None,
                 build_args=None, working_dir='.'):
        self.name = name
        self.environment = environment
        if env_sample_file is not None:
            self.env_sample_file = env_sample_file
        else:
            self.env_sample_file = './env.sample'
        self.version = version
        self.ecr_client = boto3.session.Session(region_name=self.region).client('ecr')
        self.cluster_name = get_cluster_name(environment)
        self.working_dir = working_dir
        self.build_args = build_args

    def run(self):
        log_warning("Deploying to {self.region}".format(**locals()))
        self.init_stack_info()
        if not os.path.exists(self.env_sample_file):
            raise UnrecoverableException('env.sample not found. Exiting.')
        ecr_client = EcrClient(self.name, self.region, self.build_args)
        ecr_client.set_version(self.version)
        log_intent("name: " + self.name + " | environment: " +
                   self.environment + " | version: " + str(ecr_client.version))
        log_bold("Checking image in ECR")
        ecr_client.build_and_upload_image()
        log_bold("Initiating deployment\n")

        jobs = []
        for index, service_name in enumerate(self.ecs_service_names):
            log_bold("Starting to deploy " + service_name)
            color = DEPLOYMENT_COLORS[index % 3]
            image_url = ecr_client.ecr_image_uri
            image_url += (':' + ecr_client.version)
            process = multiprocessing.Process(
                target=deployer.deploy_new_version,
                args=(
                    self.region,
                    self.cluster_name,
                    service_name,
                    ecr_client.version,
                    self.name,
                    self.env_sample_file,
                    self.environment,
                    color,
                    image_url
                )
            )
            jobs.append(process)
            process.start()

        exit_codes = []
        while True:
            sleep(1)
            exit_codes = [proc.exitcode for proc in jobs]
            if None not in exit_codes:
                break

        if any(exit_codes) != 0:
            raise UnrecoverableException("Deployment failed")

    def upload_image(self, additional_tags):
        EcrClient(self.name, self.region, self.build_args).upload_image(self.version, additional_tags)

    @property
    def region(self):
        return get_region_for_environment(self.environment)

    def init_stack_info(self):
        try:
            self.stack_name = get_service_stack_name(self.environment, self.name)
            stack = get_client_for(
                'cloudformation',
                self.environment
            ).describe_stacks(
                StackName=self.stack_name
            )['Stacks'][0]
            self.ecs_service_names = [
                service_name['OutputValue'] for service_name in list(
                    filter(
                        lambda x: x['OutputKey'].endswith('EcsServiceName'),
                        stack['Outputs']
                    )
                )
            ]
        except ClientError as client_error:
            err = str(client_error)
            if "Stack with id %s does not exist" % self.stack_name in err:
                log_err(
                    "%s cluster not found. Create the environment cluster using `create_environment` command." % self.environment)
            else:
                raise UnrecoverableException(str(client_error))
